# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib

from oslo_db import api as db_api


_BACKEND_MAPPING = {
    'sqlalchemy': 'qinling.db.sqlalchemy.api',
}

IMPL = db_api.DBAPI('sqlalchemy', backend_mapping=_BACKEND_MAPPING)


def setup_db():
    IMPL.setup_db()


def drop_db():
    IMPL.drop_db()


def start_tx():
    IMPL.start_tx()


def commit_tx():
    IMPL.commit_tx()


def rollback_tx():
    IMPL.rollback_tx()


def end_tx():
    IMPL.end_tx()


@contextlib.contextmanager
def transaction():
    with IMPL.transaction():
        yield


def delete_all():
    """A helper function for testing."""
    delete_jobs(insecure=True)
    delete_webhooks(insecure=True)
    delete_executions(insecure=True)
    delete_function_aliases(insecure=True)
    delete_functions(insecure=True)
    delete_runtimes(insecure=True)


def conditional_update(model, values, expected_values, **kwargs):
    return IMPL.conditional_update(model, values, expected_values, **kwargs)


def get_function(id, insecure=None):
    """Get function from db.

    'insecure' param is needed for job handler and webhook.
    """
    return IMPL.get_function(id, insecure=insecure)


def get_functions(limit=None, marker=None, sort_keys=None,
                  sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_functions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def create_function(values):
    return IMPL.create_function(values)


def update_function(id, values):
    return IMPL.update_function(id, values)


def delete_function(id):
    return IMPL.delete_function(id)


def delete_functions(**kwargs):
    return IMPL.delete_functions(**kwargs)


def create_runtime(values):
    return IMPL.create_runtime(values)


def get_runtime(id):
    return IMPL.get_runtime(id)


def get_runtimes():
    return IMPL.get_runtimes()


def delete_runtime(id):
    return IMPL.delete_runtime(id)


def update_runtime(id, values):
    return IMPL.update_runtime(id, values)


def delete_runtimes(**kwargs):
    return IMPL.delete_runtimes(**kwargs)


def create_execution(values):
    return IMPL.create_execution(values)


def get_execution(id):
    return IMPL.get_execution(id)


def get_executions(**filters):
    return IMPL.get_executions(**filters)


def delete_execution(id):
    return IMPL.delete_execution(id)


def update_execution(id, values):
    return IMPL.update_execution(id, values)


def delete_executions(**kwargs):
    return IMPL.delete_executions(**kwargs)


def create_job(values):
    return IMPL.create_job(values)


def get_job(id):
    return IMPL.get_job(id)


def get_next_jobs(before):
    return IMPL.get_next_jobs(before)


def delete_job(id):
    return IMPL.delete_job(id)


def update_job(id, values):
    return IMPL.update_job(id, values)


def get_jobs(**kwargs):
    return IMPL.get_jobs(**kwargs)


def delete_jobs(**kwargs):
    return IMPL.delete_jobs(**kwargs)


def create_webhook(values):
    return IMPL.create_webhook(values)


def get_webhook(id, insecure=None):
    return IMPL.get_webhook(id, insecure=insecure)


def get_webhooks(**kwargs):
    return IMPL.get_webhooks(**kwargs)


def delete_webhook(id):
    return IMPL.delete_webhook(id)


def update_webhook(id, values):
    return IMPL.update_webhook(id, values)


def delete_webhooks(**kwargs):
    return IMPL.delete_webhooks(**kwargs)


def increase_function_version(function_id, old_version, **kwargs):
    """This function is meant to be invoked within locking section."""
    return IMPL.increase_function_version(function_id, old_version, **kwargs)


def get_function_version(function_id, version, **kwargs):
    return IMPL.get_function_version(function_id, version, **kwargs)


# This function is only used in unit test.
def update_function_version(function_id, version, **kwargs):
    return IMPL.update_function_version(function_id, version, **kwargs)


def delete_function_version(function_id, version):
    return IMPL.delete_function_version(function_id, version)


def get_function_versions(**kwargs):
    return IMPL.get_function_versions(**kwargs)


def create_function_alias(**kwargs):
    return IMPL.create_function_alias(**kwargs)


def get_function_alias(name, **kwargs):
    return IMPL.get_function_alias(name, **kwargs)


def get_function_aliases(**kwargs):
    return IMPL.get_function_aliases(**kwargs)


def update_function_alias(name, **kwargs):
    return IMPL.update_function_alias(name, **kwargs)


def delete_function_alias(name, **kwargs):
    return IMPL.delete_function_alias(name, **kwargs)


# For unit test
def delete_function_aliases(**kwargs):
    return IMPL.delete_function_aliases(**kwargs)



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os

from tempest.lib import decorators
from tempest.lib import exceptions
import tenacity

from qinling_tempest_plugin.tests import base
from qinling_tempest_plugin.tests import utils


class FunctionsTest(base.BaseQinlingTest):
    name_prefix = 'FunctionsTest'

    def setUp(self):
        super(FunctionsTest, self).setUp()

        # Wait until runtime is available
        self.wait_runtime_available(self.runtime_id)
        self.python_zip_file = self.create_package()

    @decorators.idempotent_id('9c36ac64-9a44-4c44-9e44-241dcc6b0933')
    def test_crud_function(self):
        # Create function
        md5sum = utils.md5(self.python_zip_file)
        function_id = self.create_function(self.python_zip_file, md5sum=md5sum)

        # Get functions
        resp, body = self.client.get_resources('functions')
        self.assertEqual(200, resp.status)
        self.assertIn(
            function_id,
            [function['id'] for function in body['functions']]
        )

        # Download function package
        resp, data = self.client.download_function(function_id)
        self.assertEqual(200, resp.status)
        self.assertEqual('application/zip', resp['content-type'])
        self.assertEqual(os.path.getsize(self.python_zip_file), len(data))

        # Delete function
        resp = self.client.delete_resource('functions', function_id)
        self.assertEqual(204, resp.status)

    @decorators.idempotent_id('1fec41cd-b753-4cad-90c5-c89d7e710317')
    def test_create_function_md5mismatch(self):
        fake_md5 = "e807f1fcf82d132f9bb018ca6738a19f"

        with open(self.python_zip_file, 'rb') as package_data:
            resp, body = self.client.create_function(
                {"source": "package", "md5sum": fake_md5},
                self.runtime_id,
                name='test_create_function_md5mismatch',
                package_data=package_data
            )

        self.assertEqual(400, resp.status_code)

    @decorators.idempotent_id('f8dde7fc-fbcc-495c-9b39-70666b7d3f64')
    def test_get_by_admin(self):
        """test_get_by_admin

        Admin user can get the function by directly specifying the function id.
        """
        function_id = self.create_function(self.python_zip_file)

        resp, body = self.admin_client.get_function(function_id)

        self.assertEqual(200, resp.status)
        self.assertEqual(function_id, body['id'])

    @decorators.idempotent_id('051f3106-df01-4fcd-a0a3-c81c99653163')
    def test_get_all_admin(self):
        """test_get_all_admin

        Admin user needs to specify filters to get all the functions.
        """
        function_id = self.create_function(self.python_zip_file)

        resp, body = self.admin_client.get_resources('functions')

        self.assertEqual(200, resp.status)
        self.assertNotIn(
            function_id,
            [function['id'] for function in body['functions']]
        )

        resp, body = self.admin_client.get_resources(
            'functions?all_projects=true'
        )

        self.assertEqual(200, resp.status)
        self.assertIn(
            function_id,
            [function['id'] for function in body['functions']]
        )

    @decorators.idempotent_id('cd396bda-2174-4335-9f7f-2457aab61a4a')
    def test_get_all_not_allowed(self):
        # Get other projects functions by normal user
        context = self.assertRaises(
            exceptions.Forbidden,
            self.client.get_resources,
            'functions?all_projects=true'
        )
        self.assertIn(
            'Operation not allowed',
            context.resp_body.get('faultstring')
        )

    @decorators.idempotent_id('5cb44ee4-6c0c-4ede-9e6c-e1b9109eaa2c')
    def test_delete_not_allowed(self):
        """Even admin user can not delete other project's function."""
        function_id = self.create_function(self.python_zip_file)

        self.assertRaises(
            exceptions.Forbidden,
            self.admin_client.delete_resource,
            'functions',
            function_id
        )

    @decorators.idempotent_id('45df227e-3399-4412-a8d3-d40c1290bc1c')
    def test_detach(self):
        """Admin only operation."""
        function_id = self.create_function(self.python_zip_file)
        resp, _ = self.client.create_execution(
            function_id, input='{"name": "Qinling"}'
        )
        self.assertEqual(201, resp.status)

        resp, body = self.admin_client.get_function_workers(function_id)
        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['workers']))

        # Detach function
        resp, _ = self.admin_client.detach_function(function_id)
        self.assertEqual(202, resp.status)

        def _assert_workers():
            resp, body = self.admin_client.get_function_workers(function_id)
            self.assertEqual(200, resp.status)
            self.assertEqual(0, len(body['workers']))

        r = tenacity.Retrying(
            wait=tenacity.wait_fixed(1),
            stop=tenacity.stop_after_attempt(5),
            retry=tenacity.retry_if_exception_type(AssertionError)
        )
        r.call(_assert_workers)



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from multiprocessing import Process
import resource
import time


def main(number=10, **kwargs):
    soft, hard = resource.getrlimit(resource.RLIMIT_NPROC)
    print('(soft, hard): %20s %20s' % (soft, hard))

    # We set a small number inside the function to avoid being affected by
    # outside.
    resource.setrlimit(resource.RLIMIT_NPROC, (number, hard))

    processes = []
    for i in range(0, number+1):
        p = Process(
            target=_sleep,
            args=(i,)
        )
        p.start()
        processes.append(p)
    for p in processes:
        p.join()


def _sleep(index):
    time.sleep(10)



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add function aliases support

Revision ID: 004
Revises: 003
Create Date: 2018-05-17 03:09:04.888969

"""

# revision identifiers, used by Alembic.
revision = '004'
down_revision = '003'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'function_aliases',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('function_id', sa.String(length=36), nullable=False),
        sa.Column('function_version', sa.Integer, nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['function_id'], [u'functions.id']),
        sa.UniqueConstraint('function_id', 'function_version', 'project_id'),
        sa.UniqueConstraint('name', 'project_id'),
        sa.Index(
            'function_aliases_project_id_function_id_function_version',
            'project_id', 'function_id', 'function_version'
        ),
        sa.Index(
            'function_aliases_project_id_name',
            'project_id', 'name'
        )
    )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from datetime import timedelta
import threading

from futurist import periodics
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from qinling import context
from qinling.db import api as db_api
from qinling.db.sqlalchemy import models
from qinling import rpc
from qinling import status
from qinling.utils import constants
from qinling.utils import etcd_util
from qinling.utils import executions
from qinling.utils import jobs
from qinling.utils.openstack import keystone as keystone_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
_periodic_tasks = {}


@periodics.periodic(300)
def handle_function_service_expiration(ctx, engine):
    """Clean up resources related to expired functions.

    If it's image function, we will rely on the orchestrator itself to do the
    image clean up, e.g. image collection feature in kubernetes.
    """
    context.set_ctx(ctx)
    delta = timedelta(seconds=CONF.engine.function_service_expiration)
    expiry_time = datetime.utcnow() - delta

    results = db_api.get_functions(
        sort_keys=['updated_at'],
        insecure=True,
        updated_at={'lte': expiry_time}
    )

    for func_db in results:
        if not etcd_util.get_service_url(func_db.id, 0):
            continue

        LOG.info(
            'Deleting service mapping and workers for function '
            '%s(version 0)',
            func_db.id
        )

        # Delete resources related to the function
        engine.delete_function(ctx, func_db.id, 0)
        # Delete etcd keys
        etcd_util.delete_function(func_db.id, 0)

    versions = db_api.get_function_versions(
        sort_keys=['updated_at'],
        insecure=True,
        updated_at={'lte': expiry_time},
    )

    for v in versions:
        if not etcd_util.get_service_url(v.function_id, v.version_number):
            continue

        LOG.info(
            'Deleting service mapping and workers for function '
            '%s(version %s)',
            v.function_id, v.version_number
        )

        # Delete resources related to the function
        engine.delete_function(ctx, v.function_id, v.version_number)
        # Delete etcd keys
        etcd_util.delete_function(v.function_id, v.version_number)


@periodics.periodic(3)
def handle_job(engine_client):
    """Execute job task with no db transactions."""
    jobs_db = db_api.get_next_jobs(timeutils.utcnow() + timedelta(seconds=3))

    for job in jobs_db:
        job_id = job.id
        func_alias = job.function_alias

        if func_alias:
            alias = db_api.get_function_alias(func_alias, insecure=True)
            func_id = alias.function_id
            func_version = alias.function_version
        else:
            func_id = job.function_id
            func_version = job.function_version

        LOG.debug("Processing job: %s, function: %s(version %s)", job_id,
                  func_id, func_version)

        func_db = db_api.get_function(func_id, insecure=True)
        trust_id = func_db.trust_id

        try:
            # Setup context before schedule job.
            ctx = keystone_utils.create_trust_context(
                trust_id, job.project_id
            )
            context.set_ctx(ctx)

            if (job.count is not None and job.count > 0):
                job.count -= 1

            # Job delete/update is done using UPDATE ... FROM ... WHERE
            # non-locking clause.
            if job.count == 0:
                modified = db_api.conditional_update(
                    models.Job,
                    {
                        'status': status.DONE,
                        'count': 0
                    },
                    {
                        'id': job_id,
                        'status': status.RUNNING
                    },
                    insecure=True,
                )
            else:
                next_time = jobs.get_next_execution_time(
                    job.pattern,
                    job.next_execution_time
                )

                modified = db_api.conditional_update(
                    models.Job,
                    {
                        'next_execution_time': next_time,
                        'count': job.count
                    },
                    {
                        'id': job_id,
                        'next_execution_time': job.next_execution_time
                    },
                    insecure=True,
                )

            if not modified:
                LOG.warning(
                    'Job %s has been already handled by another periodic '
                    'task.', job_id
                )
                continue

            LOG.debug(
                "Starting to execute function %s(version %s) by job %s",
                func_id, func_version, job_id
            )

            params = {
                'function_id': func_id,
                'function_version': func_version,
                'input': job.function_input,
                'sync': False,
                'description': constants.EXECUTION_BY_JOB % job_id
            }
            executions.create_execution(engine_client, params)
        except Exception:
            LOG.exception("Failed to process job %s", job_id)
        finally:
            context.set_ctx(None)


def start_function_mapping_handler(engine):
    """Start function mapping handler thread.

    Function mapping handler is supposed to be running with engine service.
    """
    worker = periodics.PeriodicWorker([])
    worker.add(
        handle_function_service_expiration,
        ctx=context.Context(),
        engine=engine
    )
    _periodic_tasks[constants.PERIODIC_FUNC_MAPPING_HANDLER] = worker

    thread = threading.Thread(target=worker.start)
    thread.setDaemon(True)
    thread.start()

    LOG.info('Function mapping handler started.')


def start_job_handler():
    """Start job handler thread.

    Job handler is supposed to be running with api service.
    """
    worker = periodics.PeriodicWorker([])
    engine_client = rpc.get_engine_client()
    worker.add(
        handle_job,
        engine_client=engine_client
    )
    _periodic_tasks[constants.PERIODIC_JOB_HANDLER] = worker

    thread = threading.Thread(target=worker.start)
    thread.setDaemon(True)
    thread.start()

    LOG.info('Job handler started.')


def stop(task=None):
    if not task:
        for name, worker in _periodic_tasks.items():
            LOG.info('Stopping periodic task: %s', name)
            worker.stop()
            del _periodic_tasks[name]
    else:
        worker = _periodic_tasks.get(task)
        if worker:
            LOG.info('Stopping periodic task: %s', task)
            worker.stop()
            del _periodic_tasks[task]



# Copyright 2017 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Pike release

Revision ID: 001
Revises: None
Create Date: 2017-05-03 12:02:51.935368

"""

# revision identifiers, used by Alembic.
revision = '001'
down_revision = None

import re

from alembic import op
import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import CreateTable

from qinling.db.sqlalchemy import types as st


@compiles(CreateTable)
def _add_if_not_exists(element, compiler, **kw):
    output = compiler.visit_create_table(element, **kw)
    if element.element.info.get("check_ifexists"):
        output = re.sub(
            "^\s*CREATE TABLE", "CREATE TABLE IF NOT EXISTS", output, re.S)
    return output


def upgrade():
    op.create_table(
        'runtimes',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('image', sa.String(length=255), nullable=False),
        sa.Column('status', sa.String(length=32), nullable=False),
        sa.Column('is_public', sa.BOOLEAN, nullable=False),
        sa.PrimaryKeyConstraint('id'),
        info={"check_ifexists": True}
    )

    op.create_table(
        'functions',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('runtime_id', sa.String(length=36), nullable=True),
        sa.Column('memory_size', sa.Integer, nullable=True),
        sa.Column('timeout', sa.Integer, nullable=True),
        sa.Column('code', st.JsonLongDictType(), nullable=False),
        sa.Column('entry', sa.String(length=80), nullable=True),
        sa.Column('count', sa.Integer, nullable=False),
        sa.Column('trust_id', sa.String(length=80), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['runtime_id'], [u'runtimes.id']),
        info={"check_ifexists": True}
    )

    op.create_table(
        'executions',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('function_id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=32), nullable=False),
        sa.Column('sync', sa.BOOLEAN, nullable=False),
        sa.Column('input', st.JsonLongDictType(), nullable=True),
        sa.Column('result', st.JsonLongDictType(), nullable=True),
        sa.Column('logs', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        info={"check_ifexists": True}
    )

    op.create_table(
        'jobs',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('function_id', sa.String(length=36), nullable=False),
        sa.Column('function_input', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=32), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('pattern', sa.String(length=32), nullable=True),
        sa.Column('first_execution_time', sa.DateTime(), nullable=True),
        sa.Column('next_execution_time', sa.DateTime(), nullable=False),
        sa.Column('count', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['function_id'], [u'functions.id']),
        info={"check_ifexists": True}
    )

    op.create_table(
        'webhooks',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('function_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        info={"check_ifexists": True}
    )



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import copy
import json

from oslo_log import log as logging
import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import rpc
from qinling.utils import constants
from qinling.utils import executions
from qinling.utils.openstack import keystone as keystone_utils
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)

UPDATE_ALLOWED = set(['function_id', 'function_version', 'description',
                      'function_alias'])


class WebhooksController(rest.RestController):
    _custom_actions = {
        'invoke': ['POST'],
    }

    def __init__(self, *args, **kwargs):
        self.type = 'webhook'
        self.engine_client = rpc.get_engine_client()
        self.qinling_endpoint = keystone_utils.get_qinling_endpoint()

        super(WebhooksController, self).__init__(*args, **kwargs)

    def _add_webhook_url(self, id, webhook):
        """Add webhook_url attribute for webhook.

        We generate the url dynamically in case the service url is changing.
        """
        res = copy.deepcopy(webhook)
        url = '/'.join(
            [self.qinling_endpoint.strip('/'), constants.CURRENT_VERSION,
             'webhooks', id, 'invoke']
        )
        res.update({'webhook_url': url})
        return res

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Webhook, types.uuid)
    def get(self, id):
        LOG.info("Get %s %s.", self.type, id)
        webhook = db_api.get_webhook(id).to_dict()
        return resources.Webhook.from_dict(self._add_webhook_url(id, webhook))

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Webhooks, bool, wtypes.text)
    def get_all(self, all_projects=False, project_id=None):
        project_id, all_projects = rest_utils.get_project_params(
            project_id, all_projects
        )
        if all_projects:
            acl.enforce('webhook:get_all:all_projects', context.get_ctx())

        filters = rest_utils.get_filters(
            project_id=project_id,
        )

        LOG.info("Get all %ss. filters=%s", self.type, filters)
        webhooks = []
        for i in db_api.get_webhooks(insecure=all_projects, **filters):
            webhooks.append(
                resources.Webhook.from_dict(
                    self._add_webhook_url(i.id, i.to_dict())
                )
            )

        return resources.Webhooks(webhooks=webhooks)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Webhook,
        body=resources.Webhook,
        status_code=201
    )
    def post(self, webhook):
        acl.enforce('webhook:create', context.get_ctx())

        params = webhook.to_dict()
        if not (params.get("function_id") or params.get("function_alias")):
            raise exc.InputException(
                'Either function_alias or function_id must be provided.'
            )

        function_id = params.get('function_id', "")
        version = params.get('function_version', 0)
        function_alias = params.get('function_alias', "")

        if function_alias:
            alias_db = db_api.get_function_alias(function_alias)
            function_id = alias_db.function_id
            version = alias_db.function_version
            # If function_alias is provided, we don't store either functin id
            # or function version.
            params.update({'function_id': None,
                           'function_version': None})

        LOG.info("Creating %s, params: %s", self.type, params)

        # Even admin user can not expose normal user's function
        db_api.get_function(function_id, insecure=False)
        if version > 0:
            db_api.get_function_version(function_id, version)

        webhook_d = db_api.create_webhook(params).to_dict()

        return resources.Webhook.from_dict(
            self._add_webhook_url(webhook_d['id'], webhook_d)
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        acl.enforce('webhook:delete', context.get_ctx())
        LOG.info("Delete %s %s.", self.type, id)
        db_api.delete_webhook(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Webhook,
        types.uuid,
        body=resources.Webhook
    )
    def put(self, id, webhook):
        acl.enforce('webhook:update', context.get_ctx())

        values = {}
        for key in UPDATE_ALLOWED:
            if webhook.to_dict().get(key) is not None:
                values.update({key: webhook.to_dict()[key]})

        LOG.info('Update %s %s, params: %s', self.type, id, values)

        # Even admin user can not expose normal user's function
        webhook_db = db_api.get_webhook(id, insecure=False)
        pre_alias = webhook_db.function_alias
        pre_function_id = webhook_db.function_id
        pre_version = webhook_db.function_version

        new_alias = values.get("function_alias")
        new_function_id = values.get("function_id", pre_function_id)
        new_version = values.get("function_version", pre_version)

        function_id = pre_function_id
        version = pre_version
        if new_alias and new_alias != pre_alias:
            alias_db = db_api.get_function_alias(new_alias)
            function_id = alias_db.function_id
            version = alias_db.function_version
            # If function_alias is provided, we don't store either functin id
            # or function version.
            values.update({'function_id': None,
                           'function_version': None})
        elif new_function_id != pre_function_id or new_version != pre_version:
            function_id = new_function_id
            version = new_version
            values.update({"function_alias": None})

        db_api.get_function(function_id, insecure=False)
        if version and version > 0:
            db_api.get_function_version(function_id, version)

        webhook = db_api.update_webhook(id, values).to_dict()
        return resources.Webhook.from_dict(self._add_webhook_url(id, webhook))

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose('json')
    def invoke(self, id, **kwargs):
        with db_api.transaction():
            # The webhook url can be accessed without authentication, so
            # insecure is used here
            webhook_db = db_api.get_webhook(id, insecure=True)
            function_alias = webhook_db.function_alias

            if function_alias:
                alias = db_api.get_function_alias(function_alias,
                                                  insecure=True)
                function_id = alias.function_id
                function_version = alias.function_version
                function_db = db_api.get_function(function_id, insecure=True)
            else:
                function_db = webhook_db.function
                function_id = webhook_db.function_id
                function_version = webhook_db.function_version

            trust_id = function_db.trust_id
            project_id = function_db.project_id

        LOG.info(
            'Invoking function %s(version %s) by webhook %s',
            function_id, function_version, id
        )

        # Setup user context
        ctx = keystone_utils.create_trust_context(trust_id, project_id)
        context.set_ctx(ctx)

        params = {
            'function_id': function_id,
            'function_version': function_version,
            'sync': False,
            'input': json.dumps(kwargs),
            'description': constants.EXECUTION_BY_WEBHOOK % id
        }
        execution = executions.create_execution(self.engine_client, params)
        pecan.response.status = 202

        return {'execution_id': execution.id}



# Copyright 2017 - Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

import cotyledon
from cotyledon import oslo_config_glue
from oslo_config import cfg
from oslo_log import log as logging

from qinling import config
from qinling.engine import service as eng_service
from qinling import rpc
from qinling.utils import common

CONF = cfg.CONF


def main():
    try:
        config.parse_args(args=common.get_properly_ordered_parameters())
        common.print_server_info("engine")
        logging.setup(CONF, 'qinling')
        # Initialize RPC configuration.
        rpc.get_transport()

        sm = cotyledon.ServiceManager()
        sm.add(
            eng_service.EngineService,
            workers=1,
        )
        oslo_config_glue.setup(sm, CONF, reload_method="mutate")
        sm.run()
    except RuntimeError as excp:
        sys.stderr.write("ERROR: %s\n" % excp)
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright 2015
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

service_option = cfg.BoolOpt(
    'qinling',
    default=True,
    help="Whether or not Qinling is expected to be available"
)


qinling_group = cfg.OptGroup(name="qinling", title="Qinling Service Options")

QinlingGroup = [
    cfg.StrOpt("region",
               default="",
               help="The region name to use. If empty, the value "
                    "of identity.region is used instead. If no such region "
                    "is found in the service catalog, the first found one is "
                    "used."),
    cfg.StrOpt("catalog_type",
               default="function-engine",
               help="Catalog type of the Qinling service."),
    cfg.StrOpt('endpoint_type',
               default='publicURL',
               choices=['public', 'admin', 'internal',
                        'publicURL', 'adminURL', 'internalURL'],
               help="The endpoint type to use for the qinling service."),
    cfg.StrOpt("python_runtime_image",
               default="openstackqinling/python3-runtime:0.0.2",
               help="The Python runtime being used in the tests."),
    cfg.StrOpt("nodejs_runtime_image",
               default="openstackqinling/nodejs-runtime:0.0.1",
               help="The NodeJS runtime being used in the tests."),
    cfg.BoolOpt("allow_external_connection",
                default=False,
                help="If the tests which need external network connection "
                     "should be running."),
]



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import cotyledon
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher

from qinling.db import api as db_api
from qinling.engine import default_engine as engine
from qinling.orchestrator import base as orchestra_base
from qinling import rpc
from qinling.services import periodics
from qinling.utils.openstack import keystone as keystone_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class EngineService(cotyledon.Service):
    def __init__(self, worker_id):
        super(EngineService, self).__init__(worker_id)
        self.server = None

    def run(self):
        qinling_endpoint = keystone_utils.get_qinling_endpoint()
        orchestrator = orchestra_base.load_orchestrator(CONF, qinling_endpoint)
        db_api.setup_db()

        topic = CONF.engine.topic
        server = CONF.engine.host
        transport = messaging.get_rpc_transport(CONF)
        target = messaging.Target(topic=topic, server=server, fanout=False)
        endpoint = engine.DefaultEngine(orchestrator, qinling_endpoint)
        access_policy = dispatcher.DefaultRPCAccessPolicy
        self.server = messaging.get_rpc_server(
            transport,
            target,
            [endpoint],
            executor='threading',
            access_policy=access_policy,
            serializer=rpc.ContextSerializer(
                messaging.serializer.JsonPayloadSerializer())
        )

        LOG.info('Starting function mapping periodic task...')
        periodics.start_function_mapping_handler(endpoint)

        LOG.info('Starting engine...')
        self.server.start()

    def terminate(self):
        periodics.stop()

        if self.server:
            LOG.info('Stopping engine...')
            self.server.stop()
            self.server.wait()



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from datetime import datetime
from datetime import timedelta
import time

import mock
from oslo_config import cfg

from qinling import context
from qinling.db import api as db_api
from qinling.services import periodics
from qinling import status
from qinling.tests.unit import base

CONF = cfg.CONF


class TestPeriodics(base.DbTestCase):
    def setUp(self):
        super(TestPeriodics, self).setUp()
        self.override_config('auth_enable', False, group='pecan')

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_handle_function_service_no_function_version(self, mock_etcd_url,
                                                         mock_etcd_delete):
        db_func = self.create_function()
        function_id = db_func.id
        # Update function to simulate function execution
        db_api.update_function(function_id, {'count': 1})
        time.sleep(1.5)

        mock_etcd_url.return_value = 'http://localhost:37718'
        self.override_config('function_service_expiration', 1, 'engine')
        mock_engine = mock.Mock()

        periodics.handle_function_service_expiration(self.ctx, mock_engine)

        mock_engine.delete_function.assert_called_once_with(
            self.ctx, function_id, 0
        )
        mock_etcd_delete.assert_called_once_with(function_id, 0)

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_handle_function_service_with_function_versions(self, mock_srv_url,
                                                            mock_etcd_delete):
        db_func = self.create_function()
        function_id = db_func.id
        self.create_function_version(0, function_id, description="new_version")
        db_api.update_function_version(function_id, 1, count=1)
        time.sleep(1.5)

        self.override_config('function_service_expiration', 1, 'engine')

        # NOTE(huntxu): although we didn't create any execution using version 0
        # of the function, it is updated as a new version is created. So the
        # call to get_service_url with version 0 should return None as there is
        # not any worker for function version 0.
        def mock_srv_url_side_effect(function_id, function_version):
            return 'http://localhost:37718' if function_version != 0 else None

        mock_srv_url.side_effect = mock_srv_url_side_effect
        mock_engine = mock.Mock()

        periodics.handle_function_service_expiration(self.ctx, mock_engine)

        mock_engine.delete_function.assert_called_once_with(
            self.ctx, function_id, 1
        )
        mock_etcd_delete.assert_called_once_with(function_id, 1)

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_handle_function_service_with_versioned_function_version_0(
            self, mock_srv_url, mock_etcd_delete
    ):
        # This case tests that if a function has multiple versions, service
        # which serves executions of function version 0 is correctly handled
        # when expired.
        db_func = self.create_function()
        function_id = db_func.id
        self.create_function_version(0, function_id, description="new_version")
        # Simulate an execution using version 0
        db_api.update_function(function_id, {'count': 1})
        time.sleep(1.5)

        self.override_config('function_service_expiration', 1, 'engine')
        mock_srv_url.return_value = 'http://localhost:37718'
        mock_engine = mock.Mock()

        periodics.handle_function_service_expiration(self.ctx, mock_engine)

        mock_engine.delete_function.assert_called_once_with(
            self.ctx, function_id, 0
        )
        mock_etcd_delete.assert_called_once_with(function_id, 0)

    @mock.patch('qinling.utils.jobs.get_next_execution_time')
    def test_job_handler(self, mock_get_next):
        db_func = self.create_function()
        function_id = db_func.id

        self.assertEqual(0, db_func.count)

        now = datetime.utcnow()
        db_job = self.create_job(
            function_id=function_id,
            status=status.RUNNING,
            next_execution_time=now,
            count=2
        )
        job_id = db_job.id

        e_client = mock.Mock()
        mock_get_next.return_value = now + timedelta(seconds=1)

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        db_job = db_api.get_job(job_id)
        self.assertEqual(1, db_job.count)
        db_func = db_api.get_function(function_id)
        self.assertEqual(1, db_func.count)
        db_execs = db_api.get_executions(function_id=function_id)
        self.assertEqual(1, len(db_execs))

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        db_job = db_api.get_job(job_id)
        self.assertEqual(0, db_job.count)
        self.assertEqual(status.DONE, db_job.status)
        db_func = db_api.get_function(function_id)
        self.assertEqual(2, db_func.count)
        db_execs = db_api.get_executions(function_id=function_id)
        self.assertEqual(2, len(db_execs))

    @mock.patch('qinling.utils.jobs.get_next_execution_time')
    def test_job_handler_with_version(self, mock_next_time):
        db_func = self.create_function()
        function_id = db_func.id
        new_version = db_api.increase_function_version(function_id, 0)

        self.assertEqual(0, new_version.count)

        now = datetime.utcnow()
        db_job = self.create_job(
            function_id,
            function_version=1,
            status=status.RUNNING,
            next_execution_time=now,
            count=2
        )
        job_id = db_job.id

        e_client = mock.Mock()
        # It doesn't matter what's the returned value, but need to be in
        # datetime type.
        mock_next_time.return_value = now + timedelta(seconds=1)

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        db_job = db_api.get_job(job_id)
        self.assertEqual(1, db_job.count)
        db_func = db_api.get_function(function_id)
        self.assertEqual(0, db_func.count)
        db_version = db_api.get_function_version(function_id, 1)
        self.assertEqual(1, db_version.count)
        db_execs = db_api.get_executions(function_id=function_id,
                                         function_version=1)
        self.assertEqual(1, len(db_execs))

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        db_job = db_api.get_job(job_id)
        self.assertEqual(0, db_job.count)
        self.assertEqual(status.DONE, db_job.status)
        db_func = db_api.get_function(function_id)
        self.assertEqual(0, db_func.count)
        db_version = db_api.get_function_version(function_id, 1)
        self.assertEqual(2, db_version.count)
        db_execs = db_api.get_executions(function_id=function_id,
                                         function_version=1)
        self.assertEqual(2, len(db_execs))

    @mock.patch('qinling.utils.jobs.get_next_execution_time')
    def test_job_handler_with_alias(self, mock_next_time):
        e_client = mock.Mock()
        now = datetime.utcnow()
        # It doesn't matter what's the returned value, but need to be in
        # datetime type.
        mock_next_time.return_value = now + timedelta(seconds=1)

        # Create a alias for a function.
        alias_name = self.rand_name(name="alias", prefix=self.prefix)
        db_func = self.create_function()
        function_id = db_func.id
        db_api.create_function_alias(name=alias_name, function_id=function_id)

        self.create_job(
            function_alias=alias_name,
            status=status.RUNNING,
            next_execution_time=now,
        )

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        # Create function version 1 and update the alias.
        db_api.increase_function_version(function_id, 0)
        db_api.update_function_alias(alias_name, function_version=1)

        periodics.handle_job(e_client)
        context.set_ctx(self.ctx)

        db_func = db_api.get_function(function_id)
        self.assertEqual(1, db_func.count)
        db_version = db_api.get_function_version(function_id, 1)
        self.assertEqual(1, db_version.count)
        db_execs = db_api.get_executions(function_id=function_id)
        self.assertEqual(2, len(db_execs))



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import rpc
from qinling.utils import executions
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)


class ExecutionLogController(rest.RestController):
    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose(content_type='text/plain')
    def get_all(self, execution_id):
        LOG.info("Get logs for execution %s.", execution_id)
        execution_db = db_api.get_execution(execution_id)

        return execution_db.logs


class ExecutionsController(rest.RestController):
    log = ExecutionLogController()

    def __init__(self, *args, **kwargs):
        self.engine_client = rpc.get_engine_client()
        self.type = 'execution'

        super(ExecutionsController, self).__init__(*args, **kwargs)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Execution,
        body=resources.Execution,
        status_code=201
    )
    def post(self, body):
        ctx = context.get_ctx()
        acl.enforce('execution:create', ctx)

        params = body.to_dict()
        if not (params.get("function_id") or params.get("function_alias")):
            raise exc.InputException(
                'Either function_alias or function_id must be provided.'
            )

        LOG.info("Creating %s. [params=%s]", self.type, params)

        db_model = executions.create_execution(self.engine_client, params)

        return resources.Execution.from_db_obj(db_model)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Executions, wtypes.text, bool, wtypes.text,
                         wtypes.text, wtypes.text)
    def get_all(self, function_id=None, all_projects=False, project_id=None,
                status=None, description=None):
        """Return a list of executions.

        :param function_id: Optional. Filtering executions by function_id.
        :param project_id: Optional. Admin user can query other projects
            resources, the param is ignored for normal user.
        :param all_projects: Optional. Get resources of all projects.
        :param status: Optional. Filter by execution status.
        :param description: Optional. Filter by description.
        """
        project_id, all_projects = rest_utils.get_project_params(
            project_id, all_projects
        )
        if all_projects:
            acl.enforce('execution:get_all:all_projects', context.get_ctx())

        filters = rest_utils.get_filters(
            function_id=function_id,
            project_id=project_id,
            status=status,
            description=description
        )
        LOG.info("Get all %ss. filters=%s", self.type, filters)

        db_execs = db_api.get_executions(insecure=all_projects, **filters)
        executions = [resources.Execution.from_db_obj(db_model)
                      for db_model in db_execs]

        return resources.Executions(executions=executions)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Execution, types.uuid)
    def get(self, id):
        LOG.info("Get resource.", resource={'type': self.type, 'id': id})

        execution_db = db_api.get_execution(id)

        return resources.Execution.from_db_obj(execution_db)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        """Delete the specified Execution."""
        LOG.info("Delete resource.", resource={'type': self.type, 'id': id})

        return db_api.delete_execution(id)



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import sys

from flask import Flask
from flask import make_response
from flask import request
from oslo_concurrency import lockutils

app = Flask(__name__)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
ch.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
del app.logger.handlers[:]
app.logger.addHandler(ch)

# Deployer can specify cfs_period_us default value here.
PERIOD = 100000


def log(message, level="info"):
    global app
    log_func = getattr(app.logger, level)
    log_func(message)


@lockutils.synchronized('set_limitation', external=True,
                        lock_path='/var/lock/qinling')
def _cgroup_limit(cpu, memory_size, pid):
    """Modify 'cgroup' files to set resource limits.

    Each pod(worker) will have cgroup folders on the host cgroup filesystem,
    like '/sys/fs/cgroup/<resource_type>/kubepods/<qos_class>/pod<pod_id>/',
    to limit memory and cpu resources that can be used in pod.

    For more information about cgroup, please see [1], about sharing PID
    namespaces in kubernetes, please see also [2].

    Return None if successful otherwise a Flask.Response object.

    [1]https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/6/html/resource_management_guide/sec-creating_cgroups
    [2]https://github.com/kubernetes/kubernetes/pull/51634
    """
    hostname = os.getenv('HOSTNAME')
    pod_id = os.getenv('POD_UID')
    qos_class = None
    if os.getenv('QOS_CLASS') == 'BestEffort':
        qos_class = 'besteffort'
    elif os.getenv('QOS_CLASS') == 'Burstable':
        qos_class = 'burstable'
    elif os.getenv('QOS_CLASS') == 'Guaranteed':
        qos_class = ''

    if not pod_id or qos_class is None:
        return make_response("Failed to get current worker information", 500)

    memory_base_path = os.path.join('/qinling_cgroup', 'memory', 'kubepods',
                                    qos_class, 'pod%s' % pod_id)
    cpu_base_path = os.path.join('/qinling_cgroup', 'cpu', 'kubepods',
                                 qos_class, 'pod%s' % pod_id)
    memory_path = os.path.join(memory_base_path, hostname)
    cpu_path = os.path.join(cpu_base_path, hostname)

    if os.path.isdir(memory_base_path):
        if not os.path.isdir(memory_path):
            os.makedirs(memory_path)

    if os.path.isdir(cpu_base_path):
        if not os.path.isdir(cpu_path):
            os.makedirs(cpu_path)

    try:
        # set cpu and memory resource limits
        with open('%s/memory.limit_in_bytes' % memory_path, 'w') as f:
            f.write('%d' % int(memory_size))
        with open('%s/cpu.cfs_period_us' % cpu_path, 'w') as f:
            f.write('%d' % PERIOD)
        with open('%s/cpu.cfs_quota_us' % cpu_path, 'w') as f:
            f.write('%d' % ((int(cpu)*PERIOD/1000)))

        # add pid to 'tasks' files
        with open('%s/tasks' % memory_path, 'w') as f:
            f.write('%d' % pid)
        with open('%s/tasks' % cpu_path, 'w') as f:
            f.write('%d' % pid)
    except Exception as e:
        return make_response("Failed to modify cgroup files: %s"
                             % str(e), 500)


@app.route('/cglimit', methods=['POST'])
def cglimit():
    """Set resource limitations for execution.

    Only root user has jurisdiction to modify all cgroup files.

    :param cpu: cpu resource that execution can use in total.
    :param memory_size: RAM resource that execution can use in total.

    Currently swap ought to be disabled in kubernetes.
    """
    params = request.get_json()
    cpu = params['cpu']
    memory_size = params['memory_size']
    pid = params['pid']
    log("Set resource limits request received, params: %s" % params)

    resp = _cgroup_limit(cpu, memory_size, pid)

    return resp if resp else 'pidlimited'



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
CURRENT_VERSION = 'v1'

EXECUTION_BY_JOB = 'Created by Job %s'
EXECUTION_BY_WEBHOOK = 'Created by Webhook %s'

PERIODIC_JOB_HANDLER = 'job_handler'
PERIODIC_FUNC_MAPPING_HANDLER = 'function_mapping_handler'

PACKAGE_FUNCTION = 'package'
SWIFT_FUNCTION = 'swift'
IMAGE_FUNCTION = 'image'

MAX_PACKAGE_SIZE = 51 * 1024 * 1024

MAX_VERSION_NUMBER = 10



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools
import json

from oslo_log import log as logging
import pecan
import six
import webob
from wsme import exc as wsme_exc

from qinling import context
from qinling import exceptions as exc

LOG = logging.getLogger(__name__)

FILTER_TYPES = ('in', 'nin', 'eq', 'neq', 'gt', 'gte', 'lt', 'lte', 'has')
LIST_VALUE_FILTER_TYPES = {'in', 'nin'}


def wrap_wsme_controller_exception(func):
    """Decorator for controllers method.

    This decorator wraps controllers method to manage wsme exceptions:
    In case of expected error it aborts the request with specific status code.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exc.QinlingException as e:
            pecan.response.translatable_error = e

            LOG.error('Error during API call: %s', six.text_type(e))
            raise wsme_exc.ClientSideError(
                msg=six.text_type(e),
                status_code=e.http_code
            )

    return wrapped


def wrap_pecan_controller_exception(func):
    """Decorator for controllers method.

    This decorator wraps controllers method to manage pecan exceptions:
    In case of expected error it aborts the request with specific status code.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exc.QinlingException as e:
            LOG.error('Error during API call: %s', six.text_type(e))
            return webob.Response(
                status=e.http_code,
                content_type='application/json',
                body=json.dumps(dict(faultstring=six.text_type(e))),
                charset='UTF-8'
            )

    return wrapped


def get_filters(**params):
    """Create filters from REST request parameters.

    :param req_params: REST request parameters.
    :return: filters dictionary.
    """
    filters = {}

    for column, data in params.items():
        if data is not None:
            if isinstance(data, six.string_types):
                f_type, value = _extract_filter_type_and_value(data)
                create_or_update_filter(column, value, f_type, filters)
            else:
                create_or_update_filter(column, data, _filter=filters)

    return filters


def create_or_update_filter(column, value, filter_type='eq', _filter=None):
    """Create or Update filter.

    :param column: Column name by which user want to filter.
    :param value: Column value.
    :param filter_type: filter type. Filter type can be
                        'eq', 'neq', 'gt', 'gte', 'lte', 'in',
                        'lt', 'nin'. Default is 'eq'.
    :param _filter: Optional. If provided same filter dictionary will
                    be updated.
    :return: filter dictionary.

    """
    if _filter is None:
        _filter = {}

    _filter[column] = {filter_type: value}

    return _filter


def _extract_filter_type_and_value(data):
    """Extract filter type and its value from the data.

    :param data: REST parameter value from which filter type and
                 value can be get. It should be in format of
                 'filter_type:value'.
    :return: filter type and value.
    """
    for filter_type in FILTER_TYPES:
        prefix = filter_type + ':'
        prefix_len = len(prefix)
        if data.startswith(prefix):
            value = six.text_type(data[prefix_len:])
            if filter_type in LIST_VALUE_FILTER_TYPES:
                value = list(value.split(','))
            return filter_type, value

    # Not matching any filter types, defaults to 'eq'.
    return 'eq', data


def get_project_params(project_id, all_projects):
    ctx = context.get_ctx()

    if project_id and not ctx.is_admin:
        project_id = context.ctx().projectid
    if project_id and ctx.is_admin:
        all_projects = True

    return project_id, all_projects



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_db.sqlalchemy import models as oslo_models
import sqlalchemy as sa
from sqlalchemy.ext import declarative

from qinling import context
from qinling.utils import common


def id_column():
    return sa.Column(
        sa.String(36),
        primary_key=True,
        default=common.generate_unicode_uuid
    )


def get_project_id():
    return context.get_ctx().projectid


class _QinlingModelBase(oslo_models.ModelBase, oslo_models.TimestampMixin):
    """Base class for all Qinling SQLAlchemy DB Models."""

    __table__ = None

    __hash__ = object.__hash__

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        for col in self.__table__.columns:
            # In case of single table inheritance a class attribute
            # corresponding to a table column may not exist so we need
            # to skip these attributes.
            if (hasattr(self, col.name) and hasattr(other, col.name) and
                    getattr(self, col.name) != getattr(other, col.name)):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self):
        """sqlalchemy based automatic to_dict method."""
        d = {}

        for col in self.__table__.columns:
            d[col.name] = getattr(self, col.name)

        common.datetime_to_str(d, 'created_at')
        common.datetime_to_str(d, 'updated_at')

        return d

    def get_clone(self):
        """Clones current object, loads all fields and returns the result."""
        m = self.__class__()

        for col in self.__table__.columns:
            if hasattr(self, col.name):
                setattr(m, col.name, getattr(self, col.name))

        setattr(m, 'created_at', getattr(self, 'created_at').isoformat(' '))

        updated_at = getattr(self, 'updated_at')
        if updated_at:
            setattr(m, 'updated_at', updated_at.isoformat(' '))
        return m

    def __repr__(self):
        return '%s %s' % (type(self).__name__, self.to_dict().__repr__())


QinlingModelBase = declarative.declarative_base(cls=_QinlingModelBase)


class QinlingSecureModelBase(QinlingModelBase):
    """Base class for all secure models."""
    __abstract__ = True

    id = id_column()
    project_id = sa.Column(
        sa.String(80),
        nullable=False,
        default=get_project_id
    )



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
import pecan
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import root as v1_root

LOG = logging.getLogger(__name__)

API_STATUS = wtypes.Enum(str, 'SUPPORTED', 'CURRENT', 'DEPRECATED')


class APIVersion(resources.Resource):
    """An API Version."""

    id = wtypes.text
    "The version identifier."

    status = API_STATUS
    "The status of the API (SUPPORTED, CURRENT or DEPRECATED)."

    links = wtypes.ArrayType(resources.Link)
    "The link to the versioned API."

    @classmethod
    def sample(cls):
        return cls(
            id='v1.0',
            status='CURRENT',
            links=[
                resources.Link(target_name='v1', rel="self",
                               href='http://example.com:7070/v1')
            ]
        )


class APIVersions(resources.Resource):
    """API Versions."""
    versions = wtypes.ArrayType(APIVersion)

    @classmethod
    def sample(cls):
        v1 = APIVersion(id='v1.0', status='CURRENT', rel="self",
                        href='http://example.com:7070/v1')
        return cls(versions=[v1])


class RootController(object):
    v1 = v1_root.Controller()

    @wsme_pecan.wsexpose(APIVersions)
    def index(self):
        LOG.info("Fetching API versions.")

        host_url_v1 = '%s/%s' % (pecan.request.application_url, 'v1')
        api_v1 = APIVersion(
            id='v1.0',
            status='CURRENT',
            links=[resources.Link(href=host_url_v1, target='v1',
                                  rel="self", )]
        )

        return APIVersions(versions=[api_v1])



#    Copyright 2016 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def apply_filters(query, model, **filters):
    filter_dict = {}

    for key, value in filters.items():
        column_attr = getattr(model, key)

        if isinstance(value, dict):
            if 'in' in value:
                query = query.filter(column_attr.in_(value['in']))
            elif 'nin' in value:
                query = query.filter(~column_attr.in_(value['nin']))
            elif 'neq' in value:
                query = query.filter(column_attr != value['neq'])
            elif 'gt' in value:
                query = query.filter(column_attr > value['gt'])
            elif 'gte' in value:
                query = query.filter(column_attr >= value['gte'])
            elif 'lt' in value:
                query = query.filter(column_attr < value['lt'])
            elif 'lte' in value:
                query = query.filter(column_attr <= value['lte'])
            elif 'eq' in value:
                query = query.filter(column_attr == value['eq'])
            elif 'has' in value:
                like_pattern = '%{0}%'.format(value['has'])

                query = query.filter(column_attr.like(like_pattern))
        else:
            filter_dict[key] = value

    if filter_dict:
        query = query.filter_by(**filter_dict)

    return query



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from oslo_config import cfg

from qinling.db import api as db_api
from qinling.engine import default_engine
from qinling import exceptions as exc
from qinling import status
from qinling.tests.unit import base
from qinling.utils import common
from qinling.utils import constants


class TestDefaultEngine(base.DbTestCase):
    def setUp(self):
        super(TestDefaultEngine, self).setUp()
        self.orchestrator = mock.Mock()
        self.qinling_endpoint = 'http://127.0.0.1:7070'
        self.default_engine = default_engine.DefaultEngine(
            self.orchestrator, self.qinling_endpoint
        )
        self.rlimit = {
            'cpu': cfg.CONF.resource_limits.default_cpu,
            'memory_size': cfg.CONF.resource_limits.default_memory
        }

    def _create_running_executions(self, function_id, num):
        for _ in range(num):
            self.create_execution(function_id=function_id)

    def test_create_runtime(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id
        # Set status to verify it is changed during creation.
        db_api.update_runtime(runtime_id, {'status': status.CREATING})

        self.default_engine.create_runtime(mock.Mock(), runtime_id)

        self.orchestrator.create_pool.assert_called_once_with(
            runtime_id, runtime.image, trusted=True)

        runtime = db_api.get_runtime(runtime_id)
        self.assertEqual(status.AVAILABLE, runtime.status)

    def test_create_runtime_failed(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id
        # Set status to verify it is changed during creation.
        db_api.update_runtime(runtime_id, {'status': status.CREATING})
        self.orchestrator.create_pool.side_effect = RuntimeError

        self.default_engine.create_runtime(mock.Mock(), runtime_id)

        self.orchestrator.create_pool.assert_called_once_with(
            runtime_id, runtime.image, trusted=True)
        runtime = db_api.get_runtime(runtime_id)
        self.assertEqual(status.ERROR, runtime.status)

    def test_delete_runtime(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id

        self.default_engine.delete_runtime(mock.Mock(), runtime_id)

        self.orchestrator.delete_pool.assert_called_once_with(
            runtime_id)
        self.assertRaisesRegex(
            exc.DBEntityNotFoundError,
            "^Runtime not found \[id=%s\]$" % runtime_id,
            db_api.get_runtime, runtime_id)

    def test_update_runtime(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id
        # Set status to verify it is changed during update.
        db_api.update_runtime(runtime_id, {'status': status.UPGRADING})
        image = self.rand_name('new_image', prefix=self.prefix)
        pre_image = self.rand_name('pre_image', prefix=self.prefix)
        self.orchestrator.update_pool.return_value = True

        self.default_engine.update_runtime(
            mock.Mock(), runtime_id, image, pre_image)

        self.orchestrator.update_pool.assert_called_once_with(
            runtime_id, image=image)
        runtime = db_api.get_runtime(runtime_id)
        self.assertEqual(runtime.status, status.AVAILABLE)

    def test_update_runtime_rollbacked(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id
        # Set status to verify it is changed during update.
        db_api.update_runtime(runtime_id, {'status': status.UPGRADING})
        image = self.rand_name('new_image', prefix=self.prefix)
        pre_image = self.rand_name('pre_image', prefix=self.prefix)
        self.orchestrator.update_pool.return_value = False

        self.default_engine.update_runtime(
            mock.Mock(), runtime_id, image, pre_image)

        self.orchestrator.update_pool.assert_called_once_with(
            runtime_id, image=image)
        runtime = db_api.get_runtime(runtime_id)
        self.assertEqual(runtime.image, pre_image)
        self.assertEqual(runtime.status, status.AVAILABLE)

    @mock.patch('qinling.engine.default_engine.DefaultEngine.scaleup_function')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.utils.etcd_util.get_worker_lock')
    def test_function_load_check_no_worker(self, mock_getlock, mock_getworkers,
                                           mock_scaleup):
        function_id = common.generate_unicode_uuid()
        runtime_id = common.generate_unicode_uuid()
        lock = mock.Mock()
        lock.is_acquired.return_value = True
        mock_getlock.return_value.__enter__.return_value = lock
        mock_getworkers.return_value = []

        self.default_engine.function_load_check(function_id, 0, runtime_id)

        mock_getworkers.assert_called_once_with(function_id, 0)
        mock_scaleup.assert_called_once_with(None, function_id, 0, runtime_id,
                                             1)

    @mock.patch('qinling.engine.default_engine.DefaultEngine.scaleup_function')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.utils.etcd_util.get_worker_lock')
    def test_function_load_check_scaleup(self, mock_getlock, mock_getworkers,
                                         mock_scaleup):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        lock = mock.Mock()
        lock.is_acquired.return_value = True
        mock_getlock.return_value.__enter__.return_value = lock

        # The default concurrency is 3, we use 4 running executions against
        # 1 worker so that there will be a scaling up.
        mock_getworkers.return_value = ['worker1']
        self._create_running_executions(function_id, 4)

        self.default_engine.function_load_check(function_id, 0, runtime_id)

        mock_getworkers.assert_called_once_with(function_id, 0)
        mock_scaleup.assert_called_once_with(None, function_id, 0, runtime_id,
                                             1)

    @mock.patch('qinling.engine.default_engine.DefaultEngine.scaleup_function')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.utils.etcd_util.get_worker_lock')
    def test_function_load_check_not_scaleup(self, mock_getlock,
                                             mock_getworkers, mock_scaleup):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        lock = mock.Mock()
        lock.is_acquired.return_value = True
        mock_getlock.return_value.__enter__.return_value = lock

        # The default concurrency is 3, we use 3 running executions against
        # 1 worker so that there won't be a scaling up.
        mock_getworkers.return_value = ['worker1']
        self._create_running_executions(function_id, 3)

        self.default_engine.function_load_check(function_id, 0, runtime_id)

        mock_getworkers.assert_called_once_with(function_id, 0)
        mock_scaleup.assert_not_called()

    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.utils.etcd_util.get_worker_lock')
    def test_function_load_check_lock_wait(self, mock_getlock,
                                           mock_getworkers):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        lock = mock.Mock()
        mock_getlock.return_value.__enter__.return_value = lock
        # Lock is acquired upon the third try.
        lock.is_acquired.side_effect = [False, False, True]
        mock_getworkers.return_value = ['worker1']
        self._create_running_executions(function_id, 3)

        self.default_engine.function_load_check(function_id, 0, runtime_id)

        self.assertEqual(3, lock.is_acquired.call_count)
        mock_getworkers.assert_called_once_with(function_id, 0)

    @mock.patch('qinling.utils.etcd_util.get_worker_lock')
    def test_function_load_check_failed_to_get_worker_lock(self, mock_getlock):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        function_version = 0
        lock = mock.Mock()
        # Lock is never acquired.
        lock.is_acquired.return_value = False
        mock_getlock.return_value.__enter__.return_value = lock

        self.assertRaisesRegex(
            exc.EtcdLockException,
            "^Etcd: failed to get worker lock for function %s"
            "\(version %s\)\.$" % (function_id, function_version),
            self.default_engine.function_load_check,
            function_id, function_version, runtime_id
        )

    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_create_execution_image_type_function(self, mock_svc_url):
        """Create 2 executions for an image type function."""
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        db_api.update_function(
            function_id,
            {
                'code': {
                    'source': constants.IMAGE_FUNCTION,
                    'image': self.rand_name('image', prefix=self.prefix)
                }
            }
        )
        function = db_api.get_function(function_id)
        execution_1 = self.create_execution(function_id=function_id)
        execution_1_id = execution_1.id
        execution_2 = self.create_execution(function_id=function_id)
        execution_2_id = execution_2.id
        mock_svc_url.return_value = None
        self.orchestrator.prepare_execution.return_value = (
            mock.Mock(), None)
        self.orchestrator.run_execution.side_effect = [
            (True, {'duration': 5, 'logs': 'fake log'}),
            (False, {'duration': 0, 'output': 'Function execution failed.'})
        ]

        # Create two executions, with different results
        self.default_engine.create_execution(
            mock.Mock(), execution_1_id, function_id, 0, runtime_id
        )
        self.default_engine.create_execution(
            mock.Mock(), execution_2_id, function_id, 0, runtime_id,
            input='input'
        )

        get_service_url_calls = [
            mock.call(function_id, 0), mock.call(function_id, 0)
        ]
        mock_svc_url.assert_has_calls(get_service_url_calls)

        prepare_calls = [
            mock.call(function_id,
                      0,
                      rlimit=self.rlimit,
                      image=function.code['image'],
                      identifier=mock.ANY,
                      labels=None,
                      input=None),
            mock.call(function_id,
                      0,
                      rlimit=self.rlimit,
                      image=function.code['image'],
                      identifier=mock.ANY,
                      labels=None,
                      input='input')
        ]
        self.orchestrator.prepare_execution.assert_has_calls(prepare_calls)

        run_calls = [
            mock.call(execution_1_id,
                      function_id,
                      0,
                      rlimit=None,
                      input=None,
                      identifier=mock.ANY,
                      service_url=None,
                      entry=function.entry,
                      trust_id=function.trust_id,
                      timeout=function.timeout),
            mock.call(execution_2_id,
                      function_id,
                      0,
                      rlimit=None,
                      input='input',
                      identifier=mock.ANY,
                      service_url=None,
                      entry=function.entry,
                      trust_id=function.trust_id,
                      timeout=function.timeout)
        ]
        self.orchestrator.run_execution.assert_has_calls(run_calls)

        execution_1 = db_api.get_execution(execution_1_id)
        execution_2 = db_api.get_execution(execution_2_id)

        self.assertEqual(status.SUCCESS, execution_1.status)
        self.assertEqual('fake log', execution_1.logs)
        self.assertEqual({"duration": 5}, execution_1.result)
        self.assertEqual(status.FAILED, execution_2.status)
        self.assertEqual('', execution_2.logs)
        self.assertEqual(
            {'duration': 0, 'output': 'Function execution failed.'},
            execution_2.result
        )

    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_create_execution_prepare_execution_exception(
            self,
            etcd_util_get_service_url_mock
    ):
        """test_create_execution_prepare_execution_exception

        Create execution for image type function, prepare_execution method
        raises exception.
        """
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        db_api.update_function(
            function_id,
            {
                'code': {
                    'source': constants.IMAGE_FUNCTION,
                    'image': self.rand_name('image', prefix=self.prefix)
                }
            }
        )
        function = db_api.get_function(function_id)
        execution = self.create_execution(function_id=function_id)
        execution_id = execution.id
        prepare_execution = self.orchestrator.prepare_execution
        prepare_execution.side_effect = exc.OrchestratorException(
            'Exception in prepare_execution'
        )
        etcd_util_get_service_url_mock.return_value = None

        self.default_engine.create_execution(
            mock.Mock(), execution_id, function_id, 0, runtime_id)

        execution = db_api.get_execution(execution_id)

        self.assertEqual(status.ERROR, execution.status)
        self.assertEqual('', execution.logs)
        self.assertEqual({'output': 'Function execution failed.'},
                         execution.result)

    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_create_execution_package_type_function(
        self,
        etcd_util_get_service_url_mock
    ):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        execution = self.create_execution(function_id=function_id)
        execution_id = execution.id
        self.default_engine.function_load_check = mock.Mock(return_value='')
        etcd_util_get_service_url_mock.return_value = None
        self.orchestrator.prepare_execution.return_value = (
            mock.Mock(), 'svc_url')
        self.orchestrator.run_execution.return_value = (
            True,
            {'success': True, 'logs': 'execution log',
             'output': 'success output'})

        self.default_engine.create_execution(
            mock.Mock(), execution_id, function_id, 0, runtime_id)

        self.default_engine.function_load_check.assert_called_once_with(
            function_id, 0, runtime_id)
        etcd_util_get_service_url_mock.assert_called_once_with(function_id, 0)
        self.orchestrator.prepare_execution.assert_called_once_with(
            function_id, 0, rlimit=self.rlimit, image=None,
            identifier=runtime_id, labels={'runtime_id': runtime_id},
            input=None)
        self.orchestrator.run_execution.assert_called_once_with(
            execution_id, function_id, 0, rlimit=self.rlimit, input=None,
            identifier=runtime_id, service_url='svc_url', entry=function.entry,
            trust_id=function.trust_id, timeout=function.timeout)

        execution = db_api.get_execution(execution_id)

        self.assertEqual(execution.status, status.SUCCESS)
        self.assertEqual(execution.logs, 'execution log')
        self.assertEqual(execution.result, {'output': 'success output'})

    def test_create_execution_loadcheck_exception(self):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        execution = self.create_execution(function_id=function_id)
        execution_id = execution.id
        self.default_engine.function_load_check = mock.Mock(
            side_effect=exc.OrchestratorException(
                'Exception in scaleup_function'
            )
        )

        self.default_engine.create_execution(
            mock.Mock(), execution_id, function_id, 0, runtime_id)

        execution = db_api.get_execution(execution_id)

        self.assertEqual(status.ERROR, execution.status)
        self.assertEqual('', execution.logs)
        self.assertEqual({'output': 'Function execution failed.'},
                         execution.result)

    @mock.patch('qinling.engine.utils.get_request_data')
    @mock.patch('qinling.engine.utils.url_request')
    @mock.patch('qinling.utils.etcd_util.get_service_url')
    def test_create_execution_found_service_url(
        self,
        etcd_util_get_service_url_mock,
        engine_utils_url_request_mock,
        engine_utils_get_request_data_mock
    ):
        function = self.create_function()
        function_id = function.id
        runtime_id = function.runtime_id
        execution = self.create_execution(function_id=function_id)
        execution_id = execution.id
        self.default_engine.function_load_check = mock.Mock(return_value='')
        etcd_util_get_service_url_mock.return_value = 'svc_url'
        engine_utils_get_request_data_mock.return_value = 'data'
        engine_utils_url_request_mock.return_value = (
            False,
            {'success': False, 'logs': 'execution log',
             'output': 'failed output'})

        self.default_engine.create_execution(
            mock.Mock(), execution_id, function_id, 0, runtime_id,
            input='input')

        self.default_engine.function_load_check.assert_called_once_with(
            function_id, 0, runtime_id)
        etcd_util_get_service_url_mock.assert_called_once_with(function_id, 0)
        engine_utils_get_request_data_mock.assert_called_once_with(
            mock.ANY, function_id, 0, execution_id, self.rlimit,
            'input', function.entry, function.trust_id,
            self.qinling_endpoint, function.timeout)
        engine_utils_url_request_mock.assert_called_once_with(
            self.default_engine.session, 'svc_url/execute', body='data')

        execution = db_api.get_execution(execution_id)

        self.assertEqual(execution.status, status.FAILED)
        self.assertEqual(execution.logs, 'execution log')
        self.assertEqual(execution.result,
                         {'success': False, 'output': 'failed output'})

    def test_delete_function(self):
        function_id = common.generate_unicode_uuid()

        self.default_engine.delete_function(mock.Mock(), function_id)

        self.orchestrator.delete_function.assert_called_once_with(
            function_id, 0
        )

    @mock.patch('qinling.utils.etcd_util.create_service_url')
    @mock.patch('qinling.utils.etcd_util.create_worker')
    def test_scaleup_function(
        self,
        etcd_util_create_worker_mock,
        etcd_util_create_service_url_mock
    ):
        function_id = common.generate_unicode_uuid()
        runtime_id = common.generate_unicode_uuid()
        self.orchestrator.scaleup_function.return_value = (['worker'], 'url')

        self.default_engine.scaleup_function(
            mock.Mock(), function_id, 0, runtime_id)

        self.orchestrator.scaleup_function.assert_called_once_with(
            function_id, 0, identifier=runtime_id, count=1)
        etcd_util_create_worker_mock.assert_called_once_with(
            function_id, 'worker', version=0)
        etcd_util_create_service_url_mock.assert_called_once_with(
            function_id, 'url', version=0)

    @mock.patch('qinling.utils.etcd_util.create_service_url')
    @mock.patch('qinling.utils.etcd_util.create_worker')
    def test_scaleup_function_multiple_workers(
        self,
        etcd_util_create_worker_mock,
        etcd_util_create_service_url_mock
    ):
        function_id = common.generate_unicode_uuid()
        runtime_id = common.generate_unicode_uuid()
        self.orchestrator.scaleup_function.return_value = (
            ['worker0', 'worker1'], 'url')

        self.default_engine.scaleup_function(
            mock.Mock(), function_id, 0, runtime_id, count=2
        )

        self.orchestrator.scaleup_function.assert_called_once_with(
            function_id, 0, identifier=runtime_id, count=2
        )
        # Two new workers are created.
        expected = [mock.call(function_id, 'worker0', version=0),
                    mock.call(function_id, 'worker1', version=0)]
        etcd_util_create_worker_mock.assert_has_calls(expected)
        etcd_util_create_service_url_mock.assert_called_once_with(
            function_id, 'url', version=0
        )

    @mock.patch('qinling.utils.etcd_util.delete_worker')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    def test_scaledown_function(
        self, etcd_util_get_workers_mock, etcd_util_delete_workers_mock
    ):
        function_id = common.generate_unicode_uuid()
        etcd_util_get_workers_mock.return_value = [
            'worker_%d' % i for i in range(4)
        ]

        self.default_engine.scaledown_function(mock.Mock(), function_id)

        etcd_util_get_workers_mock.assert_called_once_with(
            function_id, 0)
        self.orchestrator.delete_worker.assert_called_once_with('worker_0')
        etcd_util_delete_workers_mock.assert_called_once_with(
            function_id, 'worker_0', version=0
        )

    @mock.patch('qinling.utils.etcd_util.delete_worker')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    def test_scaledown_function_multiple_workers(
        self, etcd_util_get_workers_mock, etcd_util_delete_workers_mock
    ):
        function_id = common.generate_unicode_uuid()
        etcd_util_get_workers_mock.return_value = [
            'worker_%d' % i for i in range(4)
        ]

        self.default_engine.scaledown_function(mock.Mock(), function_id,
                                               count=2)

        etcd_util_get_workers_mock.assert_called_once_with(function_id, 0)
        # First two workers will be deleted.
        expected = [mock.call('worker_0'), mock.call('worker_1')]
        self.orchestrator.delete_worker.assert_has_calls(expected)
        self.assertEqual(2, self.orchestrator.delete_worker.call_count)
        expected = [
            mock.call(function_id, 'worker_0', version=0),
            mock.call(function_id, 'worker_1', version=0)
        ]
        etcd_util_delete_workers_mock.assert_has_calls(expected)
        self.assertEqual(2, etcd_util_delete_workers_mock.call_count)

    @mock.patch('qinling.utils.etcd_util.delete_worker')
    @mock.patch('qinling.utils.etcd_util.get_workers')
    def test_scaledown_function_leaving_one_worker(
        self, etcd_util_get_workers_mock, etcd_util_delete_workers_mock
    ):
        function_id = common.generate_unicode_uuid()
        etcd_util_get_workers_mock.return_value = [
            'worker_%d' % i for i in range(4)
        ]

        self.default_engine.scaledown_function(
            mock.Mock(), function_id, count=5)  # count > len(workers)

        etcd_util_get_workers_mock.assert_called_once_with(function_id, 0)
        # Only the first three workers will be deleted
        expected = [
            mock.call('worker_0'), mock.call('worker_1'), mock.call('worker_2')
        ]
        self.orchestrator.delete_worker.assert_has_calls(expected)
        self.assertEqual(3, self.orchestrator.delete_worker.call_count)
        expected = [
            mock.call(function_id, 'worker_0', version=0),
            mock.call(function_id, 'worker_1', version=0),
            mock.call(function_id, 'worker_2', version=0)
        ]
        etcd_util_delete_workers_mock.assert_has_calls(expected)
        self.assertEqual(3, etcd_util_delete_workers_mock.call_count)

    def test_get_runtime_pool(self):
        runtime = self.create_runtime()
        runtime_id = runtime.id

        self.default_engine.get_runtime_pool(mock.Mock(), runtime_id)

        self.orchestrator.get_pool.assert_called_once_with(runtime_id)



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add function version support

Revision ID: 002
Revises: 001
Create Date: 2018-04-12 00:12:45.461970

"""

# revision identifiers, used by Alembic.
revision = '002'
down_revision = '001'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'function_versions',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('function_id', sa.String(length=36), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('version_number', sa.Integer, nullable=False),
        sa.Column('count', sa.Integer, nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['function_id'], [u'functions.id']),
        sa.UniqueConstraint('function_id', 'version_number', 'project_id'),
        sa.Index(
            'function_versions_project_id_function_id_version_number',
            'project_id', 'function_id', 'version_number'
        )
    )

    op.add_column(
        'functions',
        sa.Column('latest_version', sa.Integer, nullable=False),
    )

    op.add_column(
        'executions',
        sa.Column('function_version', sa.Integer, nullable=False),
    )

    op.add_column(
        'jobs',
        sa.Column('function_version', sa.Integer, nullable=False),
    )

    op.add_column(
        'webhooks',
        sa.Column('function_version', sa.Integer, nullable=False),
    )



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from concurrent import futures
import hashlib
import json
import requests

import futurist
from oslo_serialization import jsonutils
from tempest import config
from tempest.lib import decorators
from tempest.lib import exceptions
import testtools

from qinling_tempest_plugin.tests import base

CONF = config.CONF
INVOKE_ERROR = "Function execution failed because of too much resource " \
               "consumption"


class ExecutionsTest(base.BaseQinlingTest):
    name_prefix = 'ExecutionsTest'

    def setUp(self):
        super(ExecutionsTest, self).setUp()
        self.wait_runtime_available(self.runtime_id)

    @decorators.idempotent_id('2a93fab0-2dae-4748-b0d4-f06b735ff451')
    def test_crud_execution(self):
        function_id = self.create_function()
        resp, body = self.client.create_execution(function_id,
                                                  input='{"name": "Qinling"}')
        self.assertEqual(201, resp.status)
        execution_id_1 = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id_1, ignore_notfound=True)
        self.assertEqual('success', body['status'])

        # Create another execution without input
        resp, body = self.client.create_execution(function_id)
        self.assertEqual(201, resp.status)
        execution_id_2 = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id_2, ignore_notfound=True)
        self.assertEqual('success', body['status'])

        # Get executions
        resp, body = self.client.get_resources('executions')
        self.assertEqual(200, resp.status)
        expected = {execution_id_1, execution_id_2}
        actual = set([execution['id'] for execution in body['executions']])
        self.assertTrue(expected.issubset(actual))

        # Delete executions
        resp = self.client.delete_resource('executions', execution_id_1)
        self.assertEqual(204, resp.status)
        resp = self.client.delete_resource('executions', execution_id_2)
        self.assertEqual(204, resp.status)

    # @decorators.idempotent_id('6a388918-86eb-4e10-88e2-0032a7df38e9')
    # def test_create_execution_worker_lock_failed(self):
    #     """test_create_execution_worker_lock_failed
    #
    #     When creating an execution, the qinling-engine will check the load
    #     and try to scaleup the function if needed. A lock is required when
    #     doing this check.
    #
    #     In this test we acquire the lock manually, so that qinling will fail
    #     to acquire the lock.
    #     """
    #     function_id = self.create_function()
    #
    #     from qinling_tempest_plugin.tests import utils
    #     etcd3_client = utils.get_etcd_client()
    #     lock_id = "function_worker_%s_%s" % (function_id, 0)
    #     with etcd3_client.lock(id=lock_id):
    #         resp, body = self.client.create_execution(
    #             function_id, input='{"name": "Qinling"}'
    #         )
    #
    #     self.assertEqual(201, resp.status)
    #     self.assertEqual('error', body['status'])
    #     result = jsonutils.loads(body['result'])
    #     self.assertEqual('Function execution failed.', result['output'])

    @decorators.idempotent_id('2199d1e6-de7d-4345-8745-a8184d6022b1')
    def test_get_all_admin(self):
        """Admin user can get executions of other projects"""
        function_id = self.create_function()
        resp, body = self.client.create_execution(
            function_id, input='{"name": "Qinling"}'
        )
        self.assertEqual(201, resp.status)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        resp, body = self.admin_client.get_resources(
            'executions?all_projects=true'
        )
        self.assertEqual(200, resp.status)
        self.assertIn(
            execution_id,
            [execution['id'] for execution in body['executions']]
        )

    @decorators.idempotent_id('009fba47-957e-4de5-82e8-a032386d3ac0')
    def test_get_all_not_allowed(self):
        # Get other projects functions by normal user
        context = self.assertRaises(
            exceptions.Forbidden,
            self.client.get_resources,
            'executions?all_projects=true'
        )
        self.assertIn(
            'Operation not allowed',
            context.resp_body.get('faultstring')
        )

    @decorators.idempotent_id('794cdfb2-0a27-4e56-86e8-be18eee9400f')
    def test_create_with_function_version(self):
        function_id = self.create_function()
        execution_id = self.create_execution(function_id)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)

        version_1 = self.create_function_version(function_id)
        execution_id = self.create_execution(function_id, version=version_1)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)

        self.update_function_package(function_id,
                                     "python/test_python_sleep.py")
        version_2 = self.create_function_version(function_id)
        execution_id = self.create_execution(function_id, version=version_2)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertNotIn('Hello, World', body)

    @decorators.idempotent_id('dbf4bd84-bde3-4d1d-8dec-93aaf18b4b5f')
    def test_create_with_function_alias(self):
        function_id = self.create_function()

        alias_name = self.create_function_alias(function_id)
        execution_id = self.create_execution(alias_name=alias_name)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)

        version_1 = self.create_function_version(function_id)
        alias_name_1 = self.create_function_alias(function_id, version_1)
        execution_id = self.create_execution(alias_name=alias_name_1)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)

        self.update_function_package(function_id,
                                     "python/test_python_sleep.py")
        version_2 = self.create_function_version(function_id)
        alias_name_2 = self.create_function_alias(function_id, version_2)
        execution_id = self.create_execution(alias_name=alias_name_2)
        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertNotIn('Hello, World', body)

    @decorators.idempotent_id('8096cc52-64d2-4660-a657-9ac0bdd743ae')
    def test_execution_async(self):
        function_id = self.create_function()
        resp, body = self.client.create_execution(function_id, sync=False)
        self.assertEqual(201, resp.status)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        self.assertEqual('running', body['status'])
        self.wait_execution_success(execution_id)

    @decorators.idempotent_id('6cb47b1d-a8c6-48f2-a92f-c4f613c33d1c')
    def test_execution_log(self):
        function_id = self.create_function()
        resp, body = self.client.create_execution(
            function_id, input='{"name": "OpenStack"}'
        )

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('success', body['status'])

        execution_id = body['id']

        # Get execution log
        resp, body = self.client.get_execution_log(execution_id)

        self.assertEqual(200, resp.status)
        self.assertIn('Hello, OpenStack', body)

    @decorators.idempotent_id('f22097dc-37db-484d-83d3-3a97e72ec576')
    def test_execution_concurrency_no_scale(self):
        package = self.create_package(name='python/test_python_sleep.py')
        function_id = self.create_function(package_path=package)

        def _create_execution():
            resp, body = self.client.create_execution(function_id)
            return resp, body

        futs = []
        with futurist.ThreadPoolExecutor(max_workers=10) as executor:
            for _ in range(3):
                fut = executor.submit(_create_execution)
                futs.append(fut)
            for f in futures.as_completed(futs):
                # Wait until we get the response
                resp, body = f.result()

                self.assertEqual(201, resp.status)
                self.addCleanup(self.client.delete_resource, 'executions',
                                body['id'], ignore_notfound=True)
                self.assertEqual('success', body['status'])

        resp, body = self.admin_client.get_function_workers(function_id)

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['workers']))

    @decorators.idempotent_id('a5ed173a-19b7-4c92-ac78-c8862ad1d1d2')
    def test_execution_concurrency_scale_up(self):
        package = self.create_package(name='python/test_python_sleep.py')
        function_id = self.create_function(package_path=package)

        def _create_execution():
            resp, body = self.client.create_execution(function_id)
            return resp, body

        futs = []
        with futurist.ThreadPoolExecutor(max_workers=10) as executor:
            for _ in range(6):
                fut = executor.submit(_create_execution)
                futs.append(fut)
            for f in futures.as_completed(futs):
                # Wait until we get the response
                resp, body = f.result()

                self.assertEqual(201, resp.status)
                self.addCleanup(self.client.delete_resource, 'executions',
                                body['id'], ignore_notfound=True)
                self.assertEqual('success', body['status'])

        resp, body = self.admin_client.get_function_workers(function_id)
        self.assertEqual(200, resp.status)
        self.assertEqual(2, len(body['workers']))

    @decorators.idempotent_id('d0598868-e45d-11e7-9125-00224d6b7bc1')
    def test_image_function_execution(self):
        function_id = self.create_function(
            image="openstackqinling/alpine-test")
        resp, body = self.client.create_execution(function_id,
                                                  input='Qinling')

        self.assertEqual(201, resp.status)
        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        self.assertEqual('success', body['status'])
        self.assertIn('duration', jsonutils.loads(body['result']))

        resp, body = self.client.get_execution_log(execution_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Qinling', body)

    @decorators.idempotent_id('ab962144-d5b1-11e8-978f-026f8338c1e5')
    def test_image_function_execution_timeout(self):
        function_id = self.create_function(image="lingxiankong/sleep")
        resp, body = self.client.create_execution(function_id,
                                                  input='6')

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('failed', body['status'])

        result = jsonutils.loads(body['result'])

        self.assertGreaterEqual(result['duration'], 5)
        self.assertIn(
            'Function execution timeout', result['output']
        )

        # Update function timeout
        resp, _ = self.client.update_function(
            function_id,
            timeout=15
        )
        self.assertEqual(200, resp.status_code)

        resp, body = self.client.create_execution(function_id,
                                                  input='6')

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('success', body['status'])

        result = jsonutils.loads(body['result'])
        self.assertGreaterEqual(result['duration'], 6)

    @decorators.idempotent_id('ccfe67ce-e467-11e7-916c-00224d6b7bc1')
    def test_python_execution_positional_args(self):
        package = self.create_package(
            name='python/test_python_positional_args.py'
        )
        function_id = self.create_function(package_path=package)

        resp, body = self.client.create_execution(function_id,
                                                  input='Qinling')

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('success', body['status'])

        result = jsonutils.loads(body['result'])
        self.assertIn('Qinling', result['output'])

    @decorators.idempotent_id('a948382a-84af-4f0e-ad08-4297345e302c')
    def test_python_execution_file_limit(self):
        package = self.create_package(name='python/test_python_file_limit.py')
        function_id = self.create_function(package_path=package)

        resp, body = self.client.create_execution(function_id)

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('failed', body['status'])

        result = jsonutils.loads(body['result'])
        self.assertIn(
            'Too many open files', result['output']
        )

    @decorators.idempotent_id('bf6f8f35-fa88-469b-8878-7aa85a8ce5ab')
    def test_python_execution_process_number(self):
        package = self.create_package(
            name='python/test_python_process_limit.py'
        )
        function_id = self.create_function(package_path=package)

        resp, body = self.client.create_execution(function_id)

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('failed', body['status'])

        result = jsonutils.loads(body['result'])
        self.assertIn(
            'too much resource consumption', result['output']
        )

    @decorators.idempotent_id('2b5f0787-b82d-4fc4-af76-cf86d389a76b')
    def test_python_execution_memory_limit(self):
        """In this case, the following steps are taken:

        1. Create a function that requires ~80M memory to run.
        2. Create an execution using the function.
        3. Verify that the execution is killed by the OOM-killer
           because the function memory limit is only 32M(default).
        4. Increase the function memory limit to 96M.
        5. Create another execution.
        6. Check the execution finished normally.
        """

        # Create function
        package = self.create_package(
            name='python/test_python_memory_limit.py'
        )
        function_id = self.create_function(package_path=package)

        # Invoke function
        resp, body = self.client.create_execution(function_id)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        # Check the process is killed
        self.assertEqual(201, resp.status)
        result = json.loads(body['result'])
        output = result.get('output')
        self.assertEqual(INVOKE_ERROR, output)

        # Increase the memory limit to 100663296(96M).
        resp, body = self.client.update_function(
            function_id, memory_size=100663296)
        self.assertEqual(200, resp.status_code)

        # Invoke the function again
        resp, body = self.client.create_execution(function_id)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        # Check the process exited normally
        self.assertEqual(201, resp.status)
        result = json.loads(body['result'])
        output = result.get('output')
        # The function returns the length of a list containing 4 long strings.
        self.assertEqual(4, output)

    @decorators.idempotent_id('ed714f98-29fe-4e8d-b6ee-9730f92bddea')
    def test_python_execution_cpu_limit(self):
        """In this case, the following steps are taken:

        1. Create a function that takes some time to finish (calculating the
           first 50000 digits of PI)
        2. Create an execution using the function.
        3. Store the duration of the first execution.
        4. Increase the function cpu limit from 100(default) to 200 millicpu.
        5. Create another execution.
        6. Check whether the duration of the first execution is approximately
           the double of the duration of the second one as its cpu resource is
           half of the second run.
        """

        # Create function
        package = self.create_package(
            name='python/test_python_cpu_limit.py'
        )
        function_id = self.create_function(package_path=package, timeout=180)

        # Invoke function
        resp, body = self.client.create_execution(function_id)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        # Record the duration, check whether the result is correct.
        self.assertEqual(201, resp.status)
        result = json.loads(body['result'])
        output = result.get('output')
        # Only the first 15 digits are returned.
        self.assertEqual('314159265358979', output)
        first_duration = result.get('duration', 0)

        # Increase the cpu limit
        resp, body = self.client.update_function(function_id, cpu=200)
        self.assertEqual(200, resp.status_code)

        # Invoke the function again
        resp, body = self.client.create_execution(function_id)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        # Record the second duration, check whether the result is correct.
        self.assertEqual(201, resp.status)
        result = json.loads(body['result'])
        output = result.get('output')
        # Only the first 15 digits are returned.
        self.assertEqual('314159265358979', output)
        second_duration = result.get('duration', 0)

        # Check whether the duration of the first execution is approximately
        # the double (1.8x ~ 2.2x) of the duration of the second one.
        # NOTE(huntxu): on my testbed, the result is quite near 2x. However
        # it may vary in different environments, so we give a wider range
        # here.
        self.assertNotEqual(0, first_duration)
        self.assertNotEqual(0, second_duration)
        upper = second_duration * 2.5
        lower = second_duration * 1.8
        self.assertGreaterEqual(upper, first_duration)
        self.assertLessEqual(lower, first_duration)

    @decorators.idempotent_id('07edf2ff-7544-4f30-b006-fd5302a2a9cc')
    @testtools.skipUnless(CONF.qinling.allow_external_connection,
                          "External network connection is not allowed")
    def test_python_execution_public_connection(self):
        """Test connections from k8s pod to the outside.

        Create a function that reads a webpage on the Internet, to
        verify that pods in Kubernetes can connect to the outside.
        """

        # Create function
        package = self.create_package(name='python/test_python_http_get.py')
        function_id = self.create_function(package_path=package)

        url = 'https://docs.openstack.org/qinling/latest'

        # Gets the page's sha256 outside Qinling
        response = requests.get(url, timeout=10)
        page_sha256 = hashlib.sha256(response.text.encode('utf-8')).hexdigest()

        # Create an execution to get the page's sha256 with Qinling
        resp, body = self.client.create_execution(
            function_id, input='{"url": "%s"}' % url
        )
        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        self.assertEqual(201, resp.status)
        self.assertEqual('success', body['status'])
        result = json.loads(body['result'])
        self.assertEqual(page_sha256, result['output'])

    @decorators.idempotent_id('b05e3bac-b23f-11e8-9679-00224d6b7bc1')
    def test_python_execution_timeout(self):
        package = self.create_package(
            name='python/test_python_sleep.py'
        )
        function_id = self.create_function(package_path=package)

        resp, body = self.client.create_execution(
            function_id,
            input='{"seconds": 7}'
        )

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('failed', body['status'])

        result = jsonutils.loads(body['result'])

        self.assertGreaterEqual(result['duration'], 5)
        self.assertIn(
            'Function execution timeout', result['output']
        )

        # Update function timeout
        resp, _ = self.client.update_function(
            function_id,
            timeout=10
        )
        self.assertEqual(200, resp.status_code)

        resp, body = self.client.create_execution(
            function_id,
            input='{"seconds": 7}'
        )

        self.assertEqual(201, resp.status)
        self.addCleanup(self.client.delete_resource, 'executions',
                        body['id'], ignore_notfound=True)
        self.assertEqual('success', body['status'])

        result = jsonutils.loads(body['result'])
        self.assertGreaterEqual(result['duration'], 7)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import shutil
import zipfile

from oslo_log import log as logging
from oslo_utils import fileutils

from qinling import exceptions as exc
from qinling.storage import base
from qinling.utils import common

LOG = logging.getLogger(__name__)
PACKAGE_NAME_TEMPLATE = "%s_%s.zip"
# Package path name including project ID
PACKAGE_PATH_TEMPLATE = "%s/%s_%s.zip"
# Package path name including version
PACKAGE_VERSION_TEMPLATE = "%s_%s_%s.zip"


class FileSystemStorage(base.PackageStorage):
    """Interact with file system for function package storage."""

    def __init__(self, conf):
        self.base_path = conf.storage.file_system_dir

    def store(self, project_id, function, data, md5sum=None):
        """Store the function package data to local file system.

        :param project_id: Project ID.
        :param function: Function ID.
        :param data: Package file content.
        :param md5sum: The MD5 provided by the user.
        :return: A tuple (if the package is updated, MD5 value of the package)
        """
        LOG.debug(
            'Store package, function: %s, project: %s', function, project_id
        )

        project_path = os.path.join(self.base_path, project_id)
        fileutils.ensure_tree(project_path)

        # Check md5
        md5_actual = common.md5(content=data)
        if md5sum and md5_actual != md5sum:
            raise exc.InputException("Package md5 mismatch.")

        func_zip = os.path.join(
            project_path,
            PACKAGE_NAME_TEMPLATE % (function, md5_actual)
        )
        if os.path.exists(func_zip):
            return False, md5_actual

        # Save package
        new_func_zip = os.path.join(project_path, '%s.zip.new' % function)
        with open(new_func_zip, 'wb') as fd:
            fd.write(data)

        if not zipfile.is_zipfile(new_func_zip):
            fileutils.delete_if_exists(new_func_zip)
            raise exc.InputException("Package is not a valid ZIP package.")

        os.rename(new_func_zip, func_zip)

        return True, md5_actual

    def retrieve(self, project_id, function, md5sum, version=0):
        """Get function package data.

        If version is not 0, return the package data of that specific function
        version.

        :param project_id: Project ID.
        :param function: Function ID.
        :param md5sum: The function MD5.
        :param version: Optional. The function version number.
        :return: File descriptor that needs to close outside.
        """
        LOG.debug(
            'Getting package data, function: %s, version: %s, md5sum: %s, '
            'project: %s',
            function, version, md5sum, project_id
        )

        if version != 0:
            project_dir = os.path.join(self.base_path, project_id)
            for filename in os.listdir(project_dir):
                root, ext = os.path.splitext(filename)
                if (root.startswith("%s_%d" % (function, version))
                        and ext == '.zip'):
                    func_zip = os.path.join(project_dir, filename)
                    break
            else:
                raise exc.StorageNotFoundException(
                    'Package of version %d function %s for project %s not '
                    'found.' % (version, function, project_id)
                )
        else:
            func_zip = os.path.join(
                self.base_path,
                PACKAGE_PATH_TEMPLATE % (project_id, function, md5sum)
            )

        if not os.path.exists(func_zip):
            raise exc.StorageNotFoundException(
                'Package of function %s for project %s not found.' %
                (function, project_id)
            )

        f = open(func_zip, 'rb')
        LOG.debug('Found package data for function %s version %d', function,
                  version)

        return f

    def delete(self, project_id, function, md5sum, version=0):
        LOG.debug(
            'Deleting package data, function: %s, version: %s, md5sum: %s, '
            'project: %s',
            function, version, md5sum, project_id
        )

        if version != 0:
            project_dir = os.path.join(self.base_path, project_id)
            for filename in os.listdir(project_dir):
                root, ext = os.path.splitext(filename)
                if (root.startswith("%s_%d" % (function, version))
                        and ext == '.zip'):
                    func_zip = os.path.join(project_dir, filename)
                    break
            else:
                return
        else:
            func_zip = os.path.join(
                self.base_path,
                PACKAGE_PATH_TEMPLATE % (project_id, function, md5sum)
            )

        if os.path.exists(func_zip):
            os.remove(func_zip)

    def changed_since(self, project_id, function, l_md5, version):
        """Check if the function package has changed.

        Check if the function package has changed between lastest and the
        specified version.

        :param project_id: Project ID.
        :param function: Function ID.
        :param l_md5: Latest function package md5sum.
        :param version: The version number compared with.
        :return: True if changed otherwise False.
        """
        # If it's the first version creation, don't check.
        if version == 0:
            return True

        version_path = os.path.join(
            self.base_path, project_id,
            PACKAGE_VERSION_TEMPLATE % (function, version, l_md5)
        )
        if os.path.exists(version_path):
            return False

        return True

    def copy(self, project_id, function, l_md5, old_version):
        """Copy function package for a new version.

        :param project_id: Project ID.
        :param function: Function ID.
        :param l_md5: Latest function package md5sum.
        :param old_version: The version number that should copy from.
        :return: None
        """
        src_package = os.path.join(self.base_path,
                                   project_id,
                                   PACKAGE_NAME_TEMPLATE % (function, l_md5)
                                   )
        dest_package = os.path.join(self.base_path,
                                    project_id,
                                    PACKAGE_VERSION_TEMPLATE %
                                    (function, old_version + 1, l_md5))

        try:
            shutil.copyfile(src_package, dest_package)
        except Exception:
            msg = "Failed to create new function version."
            LOG.exception(msg)
            raise exc.StorageProviderException(msg)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os

import requests
from zaqarclient.queues import client


def _send_message(z_client, queue_name, status, server=''):
    queue_name = queue_name or 'test_queue'
    queue = z_client.queue(queue_name)
    queue.post({"body": {'status': status, 'server': server}})
    print 'message posted.'


def check_and_trigger(context, **kwargs):
    file_name = 'count.txt'
    r = requests.get('http://httpbin.org/status/500')

    if r.status_code != requests.codes.ok:
        if not os.path.isfile(file_name):
            count = 1
            with open(file_name, 'w') as f:
                f.write(str(count))
        else:
            with open(file_name, 'r+') as f:
                count = int(f.readline())
                count += 1
                if count == 3:
                    # Send message and stop trigger after 3 checks
                    z_client = client.Client(
                        session=context['os_session'],
                        version=2,
                    )
                    _send_message(z_client, kwargs.get('queue'), r.status_code,
                                  'api1.production.catalyst.co.nz')

                f.seek(0)
                f.write(str(count))
                f.truncate()

        print('new count: %s' % count)
    else:
        try:
            os.remove(file_name)
        except OSError:
            pass



# Copyright 2018 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tempest import config
from tempest.lib import decorators

from qinling_tempest_plugin.tests import base

CONF = config.CONF


class NodeJSExecutionsTest(base.BaseQinlingTest):
    name_prefix = 'NodeJSExecutionsTest'
    image = CONF.qinling.nodejs_runtime_image

    def setUp(self):
        super(NodeJSExecutionsTest, self).setUp()
        self.wait_runtime_available(self.runtime_id)

    @decorators.idempotent_id('e3046fa4-2289-11e8-b720-00224d6b7bc1')
    def test_basic_nodejs_execution(self):
        package = self.create_package(name='nodejs/test_nodejs_basic.js')
        function_id = self.create_function(package_path=package)
        resp, body = self.client.create_execution(function_id,
                                                  input='{"name": "Qinling"}')
        self.assertEqual(201, resp.status)
        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)
        self.assertEqual('success', body['status'])



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import importlib
import json
from multiprocessing import Manager
from multiprocessing import Process
import os
import resource
import sys
import time
import traceback

from flask import Flask
from flask import request
from flask import Response
from keystoneauth1.identity import generic
from keystoneauth1 import session
import psutil
import requests

app = Flask(__name__)

DOWNLOAD_ERROR = "Failed to download function package from %s, error: %s"
INVOKE_ERROR = "Function execution failed because of too much resource " \
               "consumption"
TIMEOUT_ERROR = "Function execution timeout."


def _print_trace():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
    print(''.join(line for line in lines))


def _set_ulimit():
    """Limit resources usage for the current process and/or its children.

    Refer to https://docs.python.org/2.7/library/resource.html
    """
    customized_limits = {
        resource.RLIMIT_NOFILE: 1024,
        resource.RLIMIT_NPROC: 128,
        # TODO(lxkong): 50M by default, need to be configurable in future.
        resource.RLIMIT_FSIZE: 524288000
    }
    for t, soft in customized_limits.items():
        _, hard = resource.getrlimit(t)
        resource.setrlimit(t, (soft, hard))


def _get_responce(output, duration, logs, success, code):
    return Response(
        response=json.dumps(
            {
                'output': output,
                'duration': duration,
                'logs': logs,
                'success': success
            }
        ),
        status=code,
        mimetype='application/json'
    )


def _killtree(pid, including_parent=True):
    parent = psutil.Process(pid)
    for child in parent.children(recursive=True):
        print("kill child %s" % child)
        child.kill()

    if including_parent:
        print("kill parent %s" % parent)
        parent.kill()


def _invoke_function(execution_id, zip_file_dir, module_name, method, arg,
                     input, return_dict, rlimit):
    """Thie function is supposed to be running in a child process.

    HOSTNAME will be used to create cgroup directory related to worker.

    Current execution pid will be added to cgroup tasks file, and then all
    its child processes will be automatically added to this 'cgroup'.

    Once executions exceed the cgroup limit, they will be killed by OOMKill
    and this subprocess will exit with number(-9).
    """
    # Set resource limit for current sub-process
    _set_ulimit()

    # Set cpu and memory limits to cgroup by calling cglimit service
    pid = os.getpid()
    root_resp = requests.post(
        'http://localhost:9092/cglimit',
        json={
            'cpu': rlimit['cpu'],
            'memory_size': rlimit['memory_size'],
            'pid': pid
        }
    )

    sys.stdout = open("%s.out" % execution_id, "w", 0)

    if not root_resp.ok:
        print('WARN: Resource limiting failed, run in unlimit mode.')

    print('Start execution: %s' % execution_id)

    sys.path.insert(0, zip_file_dir)
    try:
        module = importlib.import_module(module_name)
        func = getattr(module, method)
        return_dict['result'] = func(arg, **input) if arg else func(**input)
        return_dict['success'] = True
    except Exception as e:
        _print_trace()

        if isinstance(e, OSError) and 'Resource' in str(e):
            sys.exit(1)

        return_dict['result'] = str(e)
        return_dict['success'] = False
    finally:
        print('Finished execution: %s' % execution_id)


@app.route('/execute', methods=['POST'])
def execute():
    """Invoke function.

    Several things need to handle in this function:
    - Save the function log
    - Capture the function internal exception
    - Deal with process execution error (The process may be killed for some
      reason, e.g. unlimited memory allocation)
    - Deal with os error for process (e.g. Resource temporarily unavailable)
    """
    params = request.get_json() or {}
    input = params.get('input') or {}
    execution_id = params['execution_id']
    download_url = params.get('download_url')
    function_id = params.get('function_id')
    entry = params.get('entry')
    request_id = params.get('request_id')
    trust_id = params.get('trust_id')
    auth_url = params.get('auth_url')
    username = params.get('username')
    password = params.get('password')
    timeout = params.get('timeout')
    zip_file_dir = '/var/qinling/packages/%s' % function_id
    rlimit = {
        'cpu': params['cpu'],
        'memory_size': params['memory_size']
    }

    function_module, function_method = 'main', 'main'
    if entry:
        function_module, function_method = tuple(entry.rsplit('.', 1))

    print(
        'Request received, request_id: %s, execution_id: %s, input: %s, '
        'auth_url: %s' %
        (request_id, execution_id, input, auth_url)
    )

    ####################################################################
    #
    # Download function package by calling sidecar service. We don't check the
    # zip file existence here to avoid using partial file during downloading.
    #
    ####################################################################
    resp = requests.post(
        'http://localhost:9091/download',
        json={
            'download_url': download_url,
            'function_id': function_id,
            'token': params.get('token')
        }
    )
    if not resp.ok:
        return _get_responce(resp.content, 0, '', False, 500)

    ####################################################################
    #
    # Provide an openstack session to user's function
    #
    ####################################################################
    os_session = None
    if auth_url:
        auth = generic.Password(
            username=username,
            password=password,
            auth_url=auth_url,
            trust_id=trust_id,
            user_domain_name='Default'
        )
        os_session = session.Session(auth=auth, verify=False)
    input.update({'context': {'os_session': os_session}})

    ####################################################################
    #
    # Create a new process to run user's function
    #
    ####################################################################
    manager = Manager()
    return_dict = manager.dict()
    return_dict['success'] = False
    start = time.time()

    # Run the function in a separate process to avoid messing up the log. If
    # the timeout is reached, kill all the subprocesses.
    p = Process(
        target=_invoke_function,
        args=(execution_id, zip_file_dir, function_module, function_method,
              input.pop('__function_input', None), input, return_dict, rlimit)
    )

    timed_out = False
    p.start()
    p.join(timeout)
    if p.is_alive():
        _killtree(p.pid)
        timed_out = True

    ####################################################################
    #
    # Get execution result(log, duration, etc.)
    #
    ####################################################################
    duration = round(time.time() - start, 3)

    # Process was killed unexpectedly or finished with error.
    if p.exitcode != 0:
        output = TIMEOUT_ERROR if timed_out else INVOKE_ERROR
        success = False
    else:
        output = return_dict.get('result')
        success = return_dict['success']

    # Execution log
    with open('%s.out' % execution_id) as f:
        logs = f.read()
    os.remove('%s.out' % execution_id)

    return _get_responce(output, duration, logs, success, 200)


@app.route('/ping')
def ping():
    return 'pong'



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import pecan
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api.controllers.v1 import execution
from qinling.api.controllers.v1 import function
from qinling.api.controllers.v1 import function_alias
from qinling.api.controllers.v1 import job
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import runtime
from qinling.api.controllers.v1 import webhook


class RootResource(resources.Resource):
    """Root resource for API version 1.

    It references all other resources belonging to the API.
    """
    uri = wtypes.text


class Controller(object):
    """API root controller for version 1."""
    functions = function.FunctionsController()
    runtimes = runtime.RuntimesController()
    executions = execution.ExecutionsController()
    jobs = job.JobsController()
    webhooks = webhook.WebhooksController()
    aliases = function_alias.FunctionAliasesController()

    @wsme_pecan.wsexpose(RootResource)
    def index(self):
        return RootResource(uri='%s/%s' % (pecan.request.application_url,
                                           'v1'))



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections
import json

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
import pecan
from pecan import rest
from webob.static import FileIter
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import function_version
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import rpc
from qinling.storage import base as storage_base
from qinling.utils import common
from qinling.utils import constants
from qinling.utils import etcd_util
from qinling.utils.openstack import keystone as keystone_util
from qinling.utils.openstack import swift as swift_util
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

POST_REQUIRED = set(['code'])
CODE_SOURCE = set(['package', 'swift', 'image'])
UPDATE_ALLOWED = set(['name', 'description', 'code', 'package', 'entry',
                      'cpu', 'memory_size', 'timeout'])


class FunctionWorkerController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.FunctionWorkers, types.uuid, int)
    def get_all(self, function_id, function_version=0):
        acl.enforce('function_worker:get_all', context.get_ctx())

        LOG.info("Getting workers for function %s(version %s).", function_id,
                 function_version)

        workers = etcd_util.get_workers(function_id, version=function_version)
        workers = [
            resources.FunctionWorker.from_dict(
                {
                    'function_id': function_id,
                    'function_version': function_version,
                    'worker_name': w
                }
            ) for w in workers
        ]

        return resources.FunctionWorkers(workers=workers)


class FunctionsController(rest.RestController):
    workers = FunctionWorkerController()
    versions = function_version.FunctionVersionsController()

    _custom_actions = {
        'scale_up': ['POST'],
        'scale_down': ['POST'],
        'detach': ['POST'],
    }

    def __init__(self, *args, **kwargs):
        self.storage_provider = storage_base.load_storage_provider(CONF)
        self.engine_client = rpc.get_engine_client()

        super(FunctionsController, self).__init__(*args, **kwargs)

    def _check_swift(self, container, object):
        # Auth needs to be enabled because qinling needs to check swift
        # object using user's credential.
        if not CONF.pecan.auth_enable:
            raise exc.InputException('Swift object not supported.')

        if not swift_util.check_object(container, object):
            raise exc.InputException('Failed to validate object in Swift.')

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose(content_type='application/zip')
    @pecan.expose('json')
    def get(self, id):
        """Get function information or download function package.

        This method can support HTTP request using either
        'Accept:application/json' or no 'Accept' header.
        """
        ctx = context.get_ctx()
        acl.enforce('function:get', ctx)

        download = strutils.bool_from_string(
            pecan.request.GET.get('download', False)
        )
        func_db = db_api.get_function(id)

        if not download:
            LOG.info("Getting function %s.", id)
            pecan.override_template('json')
            return resources.Function.from_db_obj(func_db).to_dict()

        LOG.info("Downloading function %s", id)
        source = func_db.code['source']

        if source == constants.PACKAGE_FUNCTION:
            f = self.storage_provider.retrieve(func_db.project_id, id,
                                               func_db.code['md5sum'])
        elif source == constants.SWIFT_FUNCTION:
            container = func_db.code['swift']['container']
            obj = func_db.code['swift']['object']
            f = swift_util.download_object(container, obj)
        else:
            msg = 'Download image function is not allowed.'
            pecan.abort(
                status_code=405,
                detail=msg,
                headers={'Server-Error-Message': msg}
            )

        pecan.response.app_iter = (f if isinstance(f, collections.Iterable)
                                   else FileIter(f))
        pecan.response.headers['Content-Disposition'] = (
            'attachment; filename="%s"' % id
        )

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose('json')
    def post(self, **kwargs):
        # When using image to create function, runtime_id is not a required
        # param.
        if not POST_REQUIRED.issubset(set(kwargs.keys())):
            raise exc.InputException(
                'Required param is missing. Required: %s' % POST_REQUIRED
            )
        LOG.info("Creating function, params: %s", kwargs)

        values = {
            'name': kwargs.get('name'),
            'description': kwargs.get('description'),
            'runtime_id': kwargs.get('runtime_id'),
            'code': json.loads(kwargs['code']),
            'entry': kwargs.get('entry', 'main.main'),
            'cpu': kwargs.get('cpu', CONF.resource_limits.default_cpu),
            'memory_size': kwargs.get(
                'memory_size', CONF.resource_limits.default_memory
            ),
            'timeout': kwargs.get(
                'timeout', CONF.resource_limits.default_timeout
            ),
        }

        common.validate_int_in_range(
            'timeout', values['timeout'], CONF.resource_limits.min_timeout,
            CONF.resource_limits.max_timeout
        )
        common.validate_int_in_range(
            'cpu', values['cpu'], CONF.resource_limits.min_cpu,
            CONF.resource_limits.max_cpu
        )
        common.validate_int_in_range(
            'memory', values['memory_size'], CONF.resource_limits.min_memory,
            CONF.resource_limits.max_memory
        )

        source = values['code'].get('source')
        if not source or source not in CODE_SOURCE:
            raise exc.InputException(
                'Invalid code source specified, available sources: %s' %
                ', '.join(CODE_SOURCE)
            )

        if source != constants.IMAGE_FUNCTION:
            if not kwargs.get('runtime_id'):
                raise exc.InputException('"runtime_id" must be specified.')

            runtime = db_api.get_runtime(kwargs['runtime_id'])
            if runtime.status != 'available':
                raise exc.InputException(
                    'Runtime %s is not available.' % kwargs['runtime_id']
                )

        store = False
        create_trust = True
        if source == constants.PACKAGE_FUNCTION:
            store = True
            md5sum = values['code'].get('md5sum')
            data = kwargs['package'].file.read()
        elif source == constants.SWIFT_FUNCTION:
            swift_info = values['code'].get('swift', {})

            if not (swift_info.get('container') and swift_info.get('object')):
                raise exc.InputException("Both container and object must be "
                                         "provided for swift type function.")

            self._check_swift(
                swift_info.get('container'),
                swift_info.get('object')
            )
        else:
            create_trust = False
            values['entry'] = None

        if cfg.CONF.pecan.auth_enable and create_trust:
            try:
                values['trust_id'] = keystone_util.create_trust().id
                LOG.debug('Trust %s created', values['trust_id'])
            except Exception:
                raise exc.TrustFailedException(
                    'Trust creation failed for function.'
                )

        # Create function and store the package data inside a db transaction so
        # that the function won't be created if any error happened during
        # package store.
        with db_api.transaction():
            func_db = db_api.create_function(values)
            if store:
                try:
                    ctx = context.get_ctx()
                    _, actual_md5 = self.storage_provider.store(
                        ctx.projectid, func_db.id, data, md5sum=md5sum
                    )
                    values['code'].update({"md5sum": actual_md5})
                    func_db = db_api.update_function(func_db.id, values)
                except Exception as e:
                    LOG.exception("Failed to store function package.")
                    keystone_util.delete_trust(values['trust_id'])
                    raise e

        pecan.response.status = 201
        return resources.Function.from_db_obj(func_db).to_dict()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Functions, bool, wtypes.text)
    def get_all(self, all_projects=False, project_id=None):
        """Return a list of functions.

        :param project_id: Optional. Admin user can query other projects
            resources, the param is ignored for normal user.
        :param all_projects: Optional. Get resources of all projects.
        """
        project_id, all_projects = rest_utils.get_project_params(
            project_id, all_projects
        )
        if all_projects:
            acl.enforce('function:get_all:all_projects', context.get_ctx())

        filters = rest_utils.get_filters(
            project_id=project_id,
        )
        LOG.info("Get all functions. filters=%s", filters)
        db_functions = db_api.get_functions(insecure=all_projects, **filters)
        functions = [resources.Function.from_db_obj(db_model)
                     for db_model in db_functions]

        return resources.Functions(functions=functions)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        """Delete the specified function.

        Delete function will also delete all its versions.
        """
        LOG.info("Delete function %s.", id)

        with db_api.transaction():
            func_db = db_api.get_function(id)
            if len(func_db.jobs) > 0:
                raise exc.NotAllowedException(
                    'The function is still associated with running job(s).'
                )
            if len(func_db.webhooks) > 0:
                raise exc.NotAllowedException(
                    'The function is still associated with webhook(s).'
                )
            if len(func_db.aliases) > 0:
                raise exc.NotAllowedException(
                    'The function is still associated with function alias(es).'
                )

            # Even admin user can not delete other project's function because
            # the trust associated can only be removed by function owner.
            if func_db.project_id != context.get_ctx().projectid:
                raise exc.NotAllowedException(
                    'Function can only be deleted by its owner.'
                )

            # Delete trust if needed
            if func_db.trust_id:
                keystone_util.delete_trust(func_db.trust_id)

            for version_db in func_db.versions:
                # Delete all resources created by orchestrator asynchronously.
                self.engine_client.delete_function(
                    id,
                    version=version_db.version_number
                )
                # Delete etcd keys
                etcd_util.delete_function(
                    id,
                    version=version_db.version_number
                )
                # Delete function version packages. Versions is only supported
                # for package type function.
                self.storage_provider.delete(
                    func_db.project_id,
                    id,
                    None,
                    version=version_db.version_number
                )

            # Delete resources for function version 0(func_db.versions==[])
            self.engine_client.delete_function(id)
            etcd_util.delete_function(id)

            source = func_db.code['source']
            if source == constants.PACKAGE_FUNCTION:
                self.storage_provider.delete(func_db.project_id, id,
                                             func_db.code['md5sum'])

            # This will also delete function service mapping and function
            # versions as well.
            db_api.delete_function(id)

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose('json')
    def put(self, id, **kwargs):
        """Update function.

        - Function can not being used by job.
        - Function can not being executed.
        - (TODO)Function status should be changed so no execution will create
           when function is updating.
        """
        values = {}

        try:
            for key in UPDATE_ALLOWED:
                if kwargs.get(key) is not None:
                    if key == "code":
                        kwargs[key] = json.loads(kwargs[key])
                    values.update({key: kwargs[key]})
        except Exception as e:
            raise exc.InputException("Invalid input, %s" % str(e))

        LOG.info('Update function %s, params: %s', id, values)
        ctx = context.get_ctx()

        if values.get('timeout'):
            common.validate_int_in_range(
                'timeout', values['timeout'], CONF.resource_limits.min_timeout,
                CONF.resource_limits.max_timeout
            )

        db_update_only = set(['name', 'description', 'timeout'])
        if set(values.keys()).issubset(db_update_only):
            func_db = db_api.update_function(id, values)
        else:
            source = values.get('code', {}).get('source')
            md5sum = values.get('code', {}).get('md5sum')
            cpu = values.get('cpu')
            memory_size = values.get('memory_size')

            # Check cpu and memory_size values when updating.
            if cpu is not None:
                common.validate_int_in_range(
                    'cpu', values['cpu'], CONF.resource_limits.min_cpu,
                    CONF.resource_limits.max_cpu
                )
            if memory_size is not None:
                common.validate_int_in_range(
                    'memory', values['memory_size'],
                    CONF.resource_limits.min_memory,
                    CONF.resource_limits.max_memory
                )

            with db_api.transaction():
                pre_func = db_api.get_function(id)

                if len(pre_func.jobs) > 0:
                    raise exc.NotAllowedException(
                        'The function is still associated with running job(s).'
                    )

                pre_source = pre_func.code['source']
                pre_md5sum = pre_func.code.get('md5sum')

                if source and source != pre_source:
                    raise exc.InputException(
                        "The function code type can not be changed."
                    )

                if pre_source == constants.IMAGE_FUNCTION:
                    raise exc.InputException(
                        "The image type function code can not be changed."
                    )

                # Package type function. 'code' and 'entry' make sense only if
                # 'package' is provided
                package_updated = False
                if (pre_source == constants.PACKAGE_FUNCTION and
                        values.get('package') is not None):
                    if md5sum and md5sum == pre_md5sum:
                        raise exc.InputException(
                            "The function code checksum is not changed."
                        )

                    # Update the package data.
                    data = values['package'].file.read()
                    package_updated, md5sum = self.storage_provider.store(
                        ctx.projectid,
                        id,
                        data,
                        md5sum=md5sum
                    )
                    values.setdefault('code', {}).update(
                        {"md5sum": md5sum, "source": pre_source}
                    )
                    values.pop('package')

                # Swift type function
                if (pre_source == constants.SWIFT_FUNCTION and
                        "swift" in values.get('code', {})):
                    swift_info = values['code']["swift"]

                    if not (swift_info.get('container') or
                            swift_info.get('object')):
                        raise exc.InputException(
                            "Either container or object must be provided for "
                            "swift type function update."
                        )

                    new_swift_info = pre_func.code['swift']
                    new_swift_info.update(swift_info)

                    self._check_swift(
                        new_swift_info.get('container'),
                        new_swift_info.get('object')
                    )

                    values['code'] = {
                        "source": pre_source,
                        "swift": new_swift_info
                    }

                # Delete allocated resources in orchestrator and etcd.
                self.engine_client.delete_function(id)
                etcd_util.delete_function(id)

                func_db = db_api.update_function(id, values)

            # Delete the old function package if needed
            if package_updated:
                self.storage_provider.delete(ctx.projectid, id, pre_md5sum)

        pecan.response.status = 200
        return resources.Function.from_db_obj(func_db).to_dict()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        None,
        types.uuid,
        body=resources.ScaleInfo,
        status_code=202
    )
    def scale_up(self, id, scale):
        """Scale up the containers for function execution.

        This is admin only operation. The load monitoring of function execution
        depends on the monitoring solution of underlying orchestrator.
        """
        acl.enforce('function:scale_up', context.get_ctx())

        func_db = db_api.get_function(id)
        params = scale.to_dict()

        LOG.info('Starting to scale up function %s, params: %s', id, params)

        self.engine_client.scaleup_function(
            id,
            runtime_id=func_db.runtime_id,
            count=params['count']
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        None,
        types.uuid,
        body=resources.ScaleInfo,
        status_code=202
    )
    def scale_down(self, id, scale):
        """Scale down the containers for function execution.

        This is admin only operation. The load monitoring of function execution
        depends on the monitoring solution of underlying orchestrator.
        """
        acl.enforce('function:scale_down', context.get_ctx())

        db_api.get_function(id)
        workers = etcd_util.get_workers(id)
        params = scale.to_dict()
        if len(workers) <= 1:
            LOG.info('No need to scale down function %s', id)
            return

        LOG.info('Starting to scale down function %s, params: %s', id, params)
        self.engine_client.scaledown_function(id, count=params['count'])

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=202)
    def detach(self, id):
        """Detach the function from its underlying workers.

        This is admin only operation, which gives admin user a safe way to
        clean up the underlying resources allocated for the function.
        """
        acl.enforce('function:detach', context.get_ctx())

        db_api.get_function(id)
        LOG.info('Starting to detach function %s', id)

        # Delete allocated resources in orchestrator and etcd keys.
        self.engine_client.delete_function(id)
        etcd_util.delete_function(id)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import threading

from eventlet import corolocal

_th_loc_storage = threading.local()


def _get_greenlet_local_storage():
    greenlet_id = corolocal.get_ident()

    greenlet_locals = getattr(_th_loc_storage, "greenlet_locals", None)

    if not greenlet_locals:
        greenlet_locals = {}
        _th_loc_storage.greenlet_locals = greenlet_locals

    if greenlet_id in greenlet_locals:
        return greenlet_locals[greenlet_id]
    else:
        return None


def has_thread_local(var_name):
    gl_storage = _get_greenlet_local_storage()
    return gl_storage and var_name in gl_storage


def get_thread_local(var_name):
    if not has_thread_local(var_name):
        return None

    return _get_greenlet_local_storage()[var_name]


def set_thread_local(var_name, val):
    if val is None and has_thread_local(var_name):
        gl_storage = _get_greenlet_local_storage()

        # Delete variable from greenlet local storage.
        if gl_storage:
            del gl_storage[var_name]

        # Delete the entire greenlet local storage from thread local storage.
        if gl_storage and len(gl_storage) == 0:
            del _th_loc_storage.greenlet_locals[corolocal.get_ident()]

    if val is not None:
        gl_storage = _get_greenlet_local_storage()
        if not gl_storage:
            gl_storage = _th_loc_storage.greenlet_locals[
                corolocal.get_ident()] = {}

        gl_storage[var_name] = val
